# pylint: disable=R0801
"""
This module contains the button class.
"""
import tkinter as tk

from clinguin.utils.attribute_types import (
    BooleanType,
    ColorType,
    StringType,
    SymbolType,
)

from ..tkinter_utils import (
    AttributeNames,
    CallbackNames,
    ConfigureFont,
    ConfigureSize,
    ConfigureTextElementSize,
    LayoutFollower,
)
from .root_cmp import RootCmp


class Button(
    RootCmp, LayoutFollower, ConfigureSize, ConfigureFont, ConfigureTextElementSize
):
    """
    A button is a element, which is generally regarded as an active element,
    so actions are executed. For available attributes see syntax definition.
    Implementation wise it is similarly implemented as the Label and Dropdownmenu -
    to make it work for layouting, the actual button is hidden the the element is actually
    a tkinter frame (therefore self._element is a frame, whereas self._button is the button).
    """

    def __init__(self, args, cid, parent, attributes, callbacks, base_engine):
        super().__init__(args, cid, parent, attributes, callbacks, base_engine)

        self._button = None

    def _init_element(self, elements):
        button_frame = tk.Frame(elements[str(self._parent)].get_element())

        self._button = tk.Button(button_frame)
        self._configure_font_element = self._button
        self._configure_text_element_size = self._button

        return button_frame

    @classmethod
    def _get_attributes(cls, attributes=None):
        if attributes is None:
            attributes = {}

        # Label/Text
        attributes[AttributeNames.label] = {"value": "", "value_type": StringType}
        # Color
        attributes[AttributeNames.backgroundcolor] = {
            "value": "white",
            "value_type": ColorType,
        }
        attributes[AttributeNames.foregroundcolor] = {
            "value": "black",
            "value_type": ColorType,
        }
        # Interactive-Attributes
        attributes[AttributeNames.onhover] = {"value": False, "value_type": BooleanType}
        attributes[AttributeNames.onhover_background_color] = {
            "value": "white",
            "value_type": ColorType,
        }
        attributes[AttributeNames.onhover_foreground_color] = {
            "value": "black",
            "value_type": ColorType,
        }

        return attributes

    @classmethod
    def _get_callbacks(cls, callbacks=None):
        if callbacks is None:
            callbacks = {}

        callbacks[CallbackNames.click] = {
            "operation": None,
            "operation_type": SymbolType,
        }

        return callbacks

    # ----------------------------------------------------------------------------------------------
    # -----Attributes----
    # ----------------------------------------------------------------------------------------------

    def _set_label_text(self, elements):
        self._logger.debug(str(elements))
        text = self._attributes[AttributeNames.label]["value"]
        self._button.configure(text=text)

    def _set_background_color(self, elements, key=AttributeNames.backgroundcolor):
        self._logger.debug(str(elements))
        value = self._attributes[key]["value"]
        self._button.configure(background=value)

    def _set_foreground_color(self, elements, key=AttributeNames.foregroundcolor):
        self._logger.debug(str(elements))
        value = self._attributes[key]["value"]
        self._button.configure(foreground=value)

    def _set_on_hover(self, elements):
        on_hover = self._attributes[AttributeNames.onhover]["value"]
        on_hover_background_color = self._attributes[
            AttributeNames.onhover_background_color
        ]["value"]

        on_hover_foreground_color = self._attributes[
            AttributeNames.onhover_foreground_color
        ]["value"]

        if on_hover:

            def enter(event):
                self._logger.debug(str(event))
                if on_hover_background_color != "":
                    self._set_background_color(
                        elements, key=AttributeNames.onhover_background_color
                    )
                if on_hover_foreground_color != "":
                    self._set_foreground_color(
                        elements, key=AttributeNames.onhover_foreground_color
                    )

            def leave(event):
                self._logger.debug(str(event))
                self._set_background_color(elements)
                self._set_foreground_color(elements)

            self._button.bind("<Enter>", enter)
            self._button.bind("<Leave>", leave)

    # ----------------------------------------------------------------------------------------------
    # -----Actions----
    # ----------------------------------------------------------------------------------------------

    def _define_click_event(self, elements):
        self._logger.debug(str(elements))
        key = CallbackNames.click
        if self._callbacks[key] and self._callbacks[key]["operation"]:

            def click_event(event):
                self._logger.debug(str(event))
                self._base_engine.post_with_operation(self._callbacks[key]["operation"])

            self._button.bind("<Button-1>", click_event)

    def _add_component_to_elements(self, elements):
        self._button.pack(expand=True, fill="both")

        elements[str(self._id)] = self



"""
This module contains the PathType.
"""

from os import path

from .type import Type
from .utils.standard_text_processing import StandardTextProcessing


class PathType(Type):
    """
    The PathType shall be used, when a path is specified (checks if path exists).
    """

    @classmethod
    def parse(cls, parse_input: str, logger):
        parsed_string = StandardTextProcessing.parse_string_with_quotes(parse_input)

        return_value = None

        if path.exists(parsed_string):
            return_value = parsed_string
        else:
            error_string = "Could not find specified path: " + parsed_string
            logger.error(error_string)
            raise Exception(error_string)

        return return_value

    @classmethod
    def description(cls):
        return "String, can either be specified as a string or if it is simple as a symbol."



"""
module that contains all frontends
"""

from .angular_frontend.angular_frontend import AngularFrontend

__all__ = [AngularFrontend.__name__]

try:
    from .tkinter_frontend.tkinter_frontend import TkinterFrontend

    __all__ += [TkinterFrontend.__name__]
except ImportError:
    # print("------> Tkinter needs has to be installed to use the TkinterFronted")
    pass



"""
This module contains the Type class.
"""


class Type:
    """
    The Type class is the super class of all Clinguin-Value-Types, i.e. each individual type
    must be a subtype of Type.
    """

    @classmethod
    def parse(cls, parse_input: str, logger):  # pylint: disable=W0613
        """
        Every type must override this method. This method parses/checks the given value.
        """
        return None

    @classmethod
    def description(cls):
        """
        Every type must override this method. One provides a textual description of available values
        in the return string.
        """
        return ""



"""
Module that contains the PopupTypesType.
"""

from enum import auto

from .enum import EnumType
from .utils.standard_text_processing import StandardTextProcessing


class PopupTypesType(EnumType):
    """
    The PopupType is an Enum Type, which can be used to specify look and feel of the message box.
    """

    INFO = auto()
    WARNING = auto()
    ERROR = auto()

    @classmethod
    def parse(cls, parse_input: str, logger):
        parsed_string = (
            StandardTextProcessing.parse_string_with_quotes(parse_input)
        ).lower()

        return_value = None

        if parsed_string == cls.INFO.name.lower():
            return_value = cls.INFO
        elif parsed_string == cls.WARNING.name.lower():
            return_value = cls.WARNING
        elif parsed_string == cls.ERROR.name.lower() or parsed_string == "danger":
            return_value = cls.ERROR
        else:
            error_string = "Could not parse " + parsed_string + " to PopupTypesType."
            logger.error(error_string)
            raise Exception(error_string)

        return return_value

    @classmethod
    def description(cls):
        return (
            "For the popup-types three different options exists: 'info' (Default information message),"
            + "'warning' and 'error'"
        )



"""
This module contains the CallBackDefinition class.
"""


class CallBackDefinition:
    """
    Some elements use this class to define their callbacks (others just do it locally, which is the preferred method).
    """

    def __init__(self, cid, parent, click_operation, elements, callback):
        self._id = cid
        self._parent = parent
        self._click_operation = click_operation
        self._elements = elements
        self._callback = callback

    def __call__(self, *args):
        self._callback(self._id, self._parent, self._click_operation, self._elements)



"""
Module contains the FrontendOperationDto class
"""

import json


class FrontendOperationDto:
    """
    Dto class for encapsulating the json that shall be sent to the backend that handles the callbacks.
    """

    def __init__(self, function):
        self.function = function

    def to_JSON(self):  # pylint: disable=C0103
        """
        Converts DTO (self) to JSON.
        """
        return json.dumps(self, default=lambda o: o.__dict__)



"""
Util functions for generating the domain state
"""

from clingo.symbol import Function


def tag(symbols, tag_name):
    """
    Adds a predicate around all the symbols
    """
    if tag_name is None:
        return symbols
    tagged = []
    for s in symbols:
        tagged.append(Function(tag_name, [s]))
    return tagged


def solve(ctl, assumptions, on_model=lambda m: None, include_all_atoms=True):
    """
    Adds information about the browsing to the domain state.

    Includes predicate  _clinguin_unsat/0
    """
    with ctl.solve(assumptions=assumptions, yield_=True) as result:
        model_symbols = None
        for m in result:
            on_model(m)
            model_symbols = m.symbols(shown=True, atoms=include_all_atoms, theory=True)
        if model_symbols is None:
            return None, result.core()
    return model_symbols, None



"""
This module features the MenuBar class.
"""

import tkinter as tk

from .root_cmp import RootCmp


class MenuBar(RootCmp):
    """
    The MenuBar is the class which resembles the menu-bar in tkinter. There must only be one in a clinguin ui file.
    """

    def _init_element(self, elements):
        parent_element = elements[self._parent].get_element()
        menubar = tk.Menu(parent_element)
        parent_element.config(menu=menubar)

        return menubar

    def _add_component_to_elements(self, elements):
        elements[str(self._id)] = self



"""
Module contains the CallbackDto class.
"""

import json


class CallbackDto:
    """
    This class represents a Callback that is Json convertible.
    Therefore objects of this class are part of the hierarchy that will get converted to Json by the
    endpoints on sending the reply.
    """

    def __init__(self, cid, event, action, operation):
        self.id = str(cid)  # pylint: disable=C0103
        self.event = str(event)
        self.action = str(action)
        self.operation = str(operation)

    def to_JSON(self):  # pylint: disable=C0103
        """
        Converts a callback dto to json.
        """
        return json.dumps(self, default=lambda o: o.__dict__)

    def clone(self):
        """
        Creates a new CallbackDto object with the same properties.
        """
        return CallbackDto(self.id, self.event, self.action, self.operation)



"""
This module contains the ColorType class.
"""

from .type import Type
from .utils.standard_text_processing import StandardTextProcessing


class ColorType(Type):
    """
    The color type is used for specifying a color.
    """

    @classmethod
    def parse(cls, parse_input: str, logger):
        parsed_string = (
            StandardTextProcessing.parse_string_with_quotes(parse_input)
        ).lower()

        return parsed_string

    @classmethod
    def description(cls):
        return (
            "One can specify the color either by providing a simple-color symbol or string"
            + "(like white, black, etc.), or by specifying it directly via a 9-digit hex-string,"
            + "where the definition of it is: '#RRRGGGBBB' where R means Red, G means Green and B means Blue."
            + "E.g. black is '#000000000', red is '#fff000000', green '#000fff000', "
            + "blue '#000000fff' and white 'fffffffff'."
        )



import json

from reference_json_output import (
    BasicTest06,
    BasicTest07,
    BasicTest08,
    BasicTest09,
    BasicTest10,
    BasicTest11,
    BasicTest12,
)
from utils_test_utils import UtilsTestUtils


class TestBasic06_11:
    def setup_method(self, test_method):
        self.p, self.uvicorn_url = UtilsTestUtils.basic_tests_start_server(test_method)

    def teardown_method(self, test_method):
        UtilsTestUtils.shutdown_server(self.p)

    def test_basic_12(self):
        uri = f"{self.uvicorn_url}"
        received_by_postman = json.dumps(BasicTest12.get_reference_json())

        UtilsTestUtils.assert_get_request(uri, received_by_postman)

    def test_basic_11(self):
        uri = f"{self.uvicorn_url}"
        received_by_postman = json.dumps(BasicTest11.get_reference_json())

        UtilsTestUtils.assert_get_request(uri, received_by_postman)

    def test_basic_10(self):
        uri = f"{self.uvicorn_url}"
        received_by_postman = json.dumps(BasicTest10.get_reference_json())

        UtilsTestUtils.assert_get_request(uri, received_by_postman)

    def test_basic_09_order_02(self):
        uri = f"{self.uvicorn_url}"
        received_by_postman = json.dumps(BasicTest09.get_reference_json())

        UtilsTestUtils.assert_get_request(uri, received_by_postman)

        data = '{"function":"add_assumption(p(2),true)"}'
        uri = f"{self.uvicorn_url}/backend"
        received_by_postman = json.dumps(BasicTest09.post_p_1_reference_json())

        UtilsTestUtils.assert_post_request(uri, received_by_postman, data)

        uri = f"{self.uvicorn_url}"
        received_by_postman = json.dumps(BasicTest09.get_p_2_reference_json())

        UtilsTestUtils.assert_get_request(uri, received_by_postman)

        data = '{"function":"clear_assumptions"}'
        uri = f"{self.uvicorn_url}/backend"
        received_by_postman = json.dumps(BasicTest09.post_p_3_reference_json())

        UtilsTestUtils.assert_post_request(uri, received_by_postman, data)

        uri = f"{self.uvicorn_url}"
        received_by_postman = json.dumps(BasicTest09.get_reference_json())

        UtilsTestUtils.assert_get_request(uri, received_by_postman)

    def test_basic_09_order_01(self):
        uri = f"{self.uvicorn_url}"
        received_by_postman = json.dumps(BasicTest09.get_reference_json())

        UtilsTestUtils.assert_get_request(uri, received_by_postman)

    def test_basic_08(self):
        uri = f"{self.uvicorn_url}"
        received_by_postman = json.dumps(BasicTest08.get_reference_json())

        UtilsTestUtils.assert_get_request(uri, received_by_postman)

    def test_basic_07(self):
        uri = f"{self.uvicorn_url}"
        received_by_postman = json.dumps(BasicTest07.get_reference_json())

        UtilsTestUtils.assert_get_request(uri, received_by_postman)

    def test_basic_06(self):
        uri = f"{self.uvicorn_url}"
        received_by_postman = json.dumps(BasicTest06.get_reference_json())

        UtilsTestUtils.assert_get_request(uri, received_by_postman)



"""
Server module, contains core-program classes.
"""

from .application.standard_json_encoder import StandardJsonEncoder
from .data.ui_state import UIState
from .presentation.endpoints import Endpoints

__all__ = [
    Endpoints.__name__,
    StandardJsonEncoder.__name__,
    UIState.__name__,
]



# pylint: disable=R0801
"""
Module that contains the ClingraphBackend.
"""
import functools
import textwrap
from pathlib import Path

from clingo import Control
from clingo.symbol import Function, String
from clingraph import Factbase, compute_graphs
from clingraph.clingo_utils import ClingraphContext
from clorm import Raw

from clinguin.server.application.backends.clingo_backend import (
    ClingoBackend,
)
from clinguin.server.data.attribute import AttributeDao

# Self defined
from clinguin.utils import StandardTextProcessing, image_to_b64
from clinguin.utils.annotations import extends, overwrites

from ....utils.transformer import UsesSignatureTransformer


class ClingraphBackend(ClingoBackend):
    """
    Extends ClingoBackend. With this Backend it is possible to include clingraph images in the UI.
    The image is rendered based on a visualization encoding every time the UI is updated.
    Then, they are sent the client as Base64 encoding.
    """

    @extends(ClingoBackend)
    def _init_command_line(self):
        """
        Sets the arguments for computing clingraph images.
        """
        super()._init_command_line()
        # pylint: disable= attribute-defined-outside-init
        self._clingraph_files = self._args.clingraph_files
        self._select_model = self._args.select_model
        self._type = self._args.type
        self._select_graph = self._args.select_graph
        self._dir = self._args.dir
        self._name_format = self._args.name_format
        self._engine = self._args.engine

        self._intermediate_format = self._args.intermediate_format

        self._attribute_image_key = "image_type"
        self._attribute_image_value = "clingraph"

    # ---------------------------------------------
    # Class methods
    # ---------------------------------------------

    @classmethod
    def register_options(cls, parser):
        """
        Registers command line options for ClingraphBackend.
        """
        ClingoBackend.register_options(parser)

        parser.add_argument(
            "--clingraph-files",
            help=textwrap.dedent(
                """\
                            A visualization encoding that will be used to generate the graph facts
                            by calling clingo with the input.
                            This encoding is expected to have only one stable model."""
            ),
            # type=argparse.FileType('r'),
            nargs="+",
            metavar="",
        )

        parser.add_argument(
            "--prefix",
            default="",
            help=textwrap.dedent(
                """\
                            Prefix expected in all the considered facts.
                            Example: --prefix=viz_ will look for predicates named viz_node, viz_edge etc."""
            ),
            type=str,
            metavar="",
        )

        parser.add_argument(
            "--default-graph",
            default="default",
            help=textwrap.dedent(
                """\
                        The name of the default graph.
                        All nodes and edges with arity 1 will be assigned to this graph
                            (default: %(default)s)"""
            ),
            type=str,
            metavar="",
        )

        parser.add_argument(
            "--select-graph",
            help=textwrap.dedent(
                """\
                    Select one of the graphs by name.
                    Can appear multiple times to select multiple graphs"""
            ),
            type=str,
            action="append",
            nargs="?",
            metavar="",
        )

        parser.add_argument(
            "--select-model",
            help=textwrap.dedent(
                """\
                    Select only one of the models when using a json input.
                    Defined by a number starting in index 0.
                    Can appear multiple times to select multiple models."""
            ),
            type=int,
            action="append",
            nargs="?",
            metavar="",
        )

        parser.add_argument(
            "--name-format",
            help=textwrap.dedent(
                """\
                        An optional string to format the name when saving multiple files,
                        this string can reference parameters `{graph_name}` and `{model_number}`.
                        Example `new_version-{graph_name}-{model_number}`
                        (default: `{graph_name}` or
                                `{model_number}/{graph_name}` for multi model functionality from json)"""
            ),
            type=str,
            metavar="",
        )

        parser.add_argument(
            "--dir",
            default="out",
            help=textwrap.dedent(
                """\
                            Directory for writing the output files
                                (default: %(default)s)"""
            ),
            type=str,
            metavar="",
        )

        parser.add_argument(
            "--type",
            default="graph",
            choices=["graph", "digraph"],
            help=textwrap.dedent(
                """\
                        The type of the graph
                        {graph|digraph}
                            (default: %(default)s)"""
            ),
            type=str,
            metavar="",
        )

        parser.add_argument(
            "--engine",
            default="dot",
            choices=[
                "dot",
                "neato",
                "twopi",
                "circo",
                "fdp",
                "osage",
                "patchwork",
                "sfdp",
            ],
            help=textwrap.dedent(
                """\
                            Layout command used by graphviz
                            {dot|neato|twopi|circo|fdp|osage|patchwork|sfdp}
                                (default: %(default)s)"""
            ),
            type=str,
            metavar="",
        )

        parser.add_argument(
            "--intermediate-format",
            default="svg",
            type=str,
            choices=[
                "png",
                "svg",
            ],
            help="Intermediate format. Use 'svg' for angular fronted and 'png' tkinter. (default: %(default)s)",
        )

    @functools.lru_cache(maxsize=None)  # pylint: disable=[method-cache-max-size-none]
    @overwrites(ClingoBackend)
    def _ui_uses_predicate(self, name: str, arity: int):
        """
        Returns a truth value of weather the ui_files contain the given signature.

        Args:
            name (str): Predicate name
            arity (int): Predicate arity
        """
        transformer = UsesSignatureTransformer(name, arity)
        self._logger.debug("Transformer parsing UI files to find %s/%s", name, arity)
        transformer.parse_files(self._ui_files + self._clingraph_files)
        return transformer.contained

    # ---------------------------------------------
    # UI update
    # ---------------------------------------------

    @extends(ClingoBackend)
    def _update_ui_state(self):
        """
        Updates the UI state by calling all domain state methods
        and creating a new control object (ui_control) using the ui_files provided
        """
        super()._update_ui_state()
        domain_state = self._domain_state
        graphs = self._compute_clingraph_graphs(domain_state)
        self._replace_uifb_with_b64_images_clingraph(graphs)

    # ---------------------------------------------
    # Private methods
    # ---------------------------------------------

    def _compute_clingraph_graphs(self, domain_state):
        """
        Computes all the graphs using the encoding and the domain state

        Arguments:

            domain_state (str): The model, brave, and cautious consequences (domain-state)
        """
        fbs = []
        ctl = Control("0")

        for f in self._clingraph_files:
            path = Path(f)
            if not path.is_file():
                self._logger.critical("File %s does not exist", f)
                raise Exception(f"File {f} does not exist")
            try:
                ctl.load(str(f))
            except Exception as e:
                self._logger.critical(
                    "Failed to load file %s (there is likely a syntax error in this logic program file).",
                    f,
                )
                raise e

        ctl.add("base", [], domain_state)
        ctl.ground([("base", [])], ClingraphContext())

        ctl.solve(on_model=lambda m: fbs.append(Factbase.from_model(m)))
        if self._select_model is not None:
            for m in self._select_model:
                if m >= len(fbs):
                    raise ValueError(f"Invalid model number selected {m}")
            fbs = [f if i in self._select_model else None for i, f in enumerate(fbs)]

        if len(fbs) > 1:
            self._logger.warning(
                "Multiple clingraph outputs were computed. Only first one considered."
            )
        graphs = compute_graphs([fbs[0]], graphviz_type=self._type)
        return graphs

    def _replace_uifb_with_b64_images_clingraph(self, graphs):
        """
        Replaces the clingraph predicates of the UI with the computed graphs.

        Arguments:
            graphs (dic) The computed graphs
        """
        attributes = list(self._ui_state.get_attributes(key=self._attribute_image_key))
        for attribute in attributes:
            attribute_value = StandardTextProcessing.parse_string_with_quotes(
                str(attribute.value)
            )
            is_cg_image = attribute_value.startswith(self._attribute_image_value)

            if not is_cg_image:
                continue

            graph_name = "default"
            split = attribute_value.split("__")
            if len(split) > 1:
                graph_name = split[1]

            image_value = self._create_image_from_graph(graphs, key=graph_name)
            new_image_key = "image"
            base64_key_image = image_to_b64(image_value)

            new_attribute = AttributeDao(
                Raw(Function(str(attribute.id), [])),
                Raw(Function(str(new_image_key), [])),
                Raw(String(str(base64_key_image))),
            )
            self._ui_state.add_attribute_direct(new_attribute)

    def _create_image_from_graph(self, graphs, position=None, key=None):
        """
        Creates the image of the graph using clingraph

        Arguments:
            graphs (dic) The computed graphs
            position (int) The position of the graph to show
            key (int) The key of the graph to show
        """
        graphs = graphs[0]
        if position is not None:
            if (len(graphs) - 1) >= position:
                graph = graphs[list(graphs.keys())[position]]
            else:
                self._logger.error("Attempted to access not valid position")
                raise Exception("Attempted to access not valid position")
        elif key is not None:
            if key in graphs:
                graph = graphs[key]
            else:
                self._logger.error("Key not found in graphs: %s", str(key))
                raise Exception("Key not found in graphs: " + str(key))
        else:
            self._logger.error("Must either specify position or key!")
            raise Exception("Must either specify position or key!")

        graph.format = self._intermediate_format
        img = graph.pipe(engine=self._engine)
        return img



"""
Contains the tkinter-gui class.
"""

import networkx as nx

from clinguin.client import AbstractFrontend
from clinguin.show_frontend_syntax_enum import ShowFrontendSyntaxEnum
from clinguin.utils.attribute_types.child_layout import ChildLayoutType
from clinguin.utils.logger import colored_text

from .tkinter_elements import (
    Button,
    Canvas,
    Container,
    Dropdownmenu,
    DropdownmenuItem,
    Label,
    MenuBar,
    MenuBarSection,
    MenuBarSectionItem,
    Message,
    RootCmp,
    Window,
)
from .tkinter_utils import AttributeNames, CallbackNames


class TkinterFrontend(AbstractFrontend):
    """
    Class that inherits from AbstractFrontend and is therefore a dynamically loaded class,
    if it shall be used to render the Frontend. It defines what to do for each element.
    """

    def __init__(self, base_engine, args):
        super().__init__(base_engine, args)

        self.elements = {}
        self.first = True

    @classmethod
    def register_options(cls, parser):
        parser.description = (
            "This GUI is based on the Python-Tkinter package and uses tkinter elements."
        )

    @classmethod
    def available_syntax(cls, show_level):
        def append_dict(description, d, type_name):
            for key in d.keys():
                description += colored_text("    " + key + "\n", "GREEN")
                if show_level == ShowFrontendSyntaxEnum.FULL:
                    if "description" in d[key]:
                        # Specific has higher priority
                        description = description + "      Description: "
                        description = description + ": " + d[key]["description"]
                        description = description + "\n"
                    elif key in AttributeNames.descriptions:
                        # General lesser priority
                        description = description + "      Description: "
                        description = (
                            description
                            + ": "
                            + colored_text(AttributeNames.descriptions[key], "MAGENTA")
                        )
                        description = description + "\n"
                    elif key in CallbackNames.descriptions:
                        description = description + "      Description: "
                        description = (
                            description
                            + ": "
                            + colored_text(CallbackNames.descriptions[key], "MAGENTA")
                        )
                        description = description + "\n"

                    if type_name in d[key]:
                        description = description + "      Possible-Values: "
                        description = description + colored_text(
                            d[key][type_name].description() + "\n", "MAGENTA"
                        )

            return description

        description = (
            "Here one finds the supported attributes and events of the TkinterFrontend "
            + "The following list shows for each the possible attributes and events."
            + "See the documentation for more details on the syntax.\n"
        )

        class_list = RootCmp.__subclasses__()

        description = description + "--------------------------------\n"
        for c in class_list:
            description = description + colored_text("" + c.__name__ + "\n", "BLUE")

            attributes = c.get_attributes()
            if len(attributes.keys()) > 0:
                description = description + colored_text("  attributes\n", "YELLOW")
                description = append_dict(description, attributes, "value_type")

            callbacks = c.get_callbacks()
            if len(callbacks.keys()) > 0:
                description = description + colored_text("  events\n", "YELLOW")
                description = append_dict(description, callbacks, "operation_type")
            description = description + "--------------------------------\n"

        return description

    def window(self, cid, parent, attributes, callbacks):
        if self.first:
            window = Window(
                self._args, cid, parent, attributes, callbacks, self._base_engine
            )
            window.add_component(self.elements)
        else:
            keys = list(self.elements.keys()).copy()
            for key in keys:
                if str(key) == str(cid):
                    continue

                if str(key) in self.elements:
                    self.elements[str(key)].forget_children(self.elements)
                    del self.elements[str(key)]

    def container(self, cid, parent, attributes, callbacks):
        container = Container(
            self._args, cid, parent, attributes, callbacks, self._base_engine
        )
        container.add_component(self.elements)

    def dropdown_menu(self, cid, parent, attributes, callbacks):
        menu = Dropdownmenu(
            self._args, cid, parent, attributes, callbacks, self._base_engine
        )
        menu.add_component(self.elements)

    def dropdown_menu_item(self, cid, parent, attributes, callbacks):
        menu = DropdownmenuItem(
            self._args, cid, parent, attributes, callbacks, self._base_engine
        )
        menu.add_component(self.elements)

    def label(self, cid, parent, attributes, callbacks):
        label = Label(self._args, cid, parent, attributes, callbacks, self._base_engine)
        label.add_component(self.elements)

    def button(self, cid, parent, attributes, callbacks):
        button = Button(
            self._args, cid, parent, attributes, callbacks, self._base_engine
        )
        button.add_component(self.elements)

    def menu_bar(self, cid, parent, attributes, callbacks):
        menubar = MenuBar(
            self._args, cid, parent, attributes, callbacks, self._base_engine
        )
        menubar.add_component(self.elements)

    def menu_bar_section(self, cid, parent, attributes, callbacks):
        menubar = MenuBarSection(
            self._args, cid, parent, attributes, callbacks, self._base_engine
        )
        menubar.add_component(self.elements)

    def menu_bar_section_item(self, cid, parent, attributes, callbacks):
        menubar = MenuBarSectionItem(
            self._args, cid, parent, attributes, callbacks, self._base_engine
        )
        menubar.add_component(self.elements)

    def message(self, cid, parent, attributes, callbacks):
        message = Message(
            self._args, cid, parent, attributes, callbacks, self._base_engine
        )
        message.add_component(self.elements)

    def canvas(self, cid, parent, attributes, callbacks):
        canvas = Canvas(
            self._args, cid, parent, attributes, callbacks, self._base_engine
        )
        canvas.add_component(self.elements)

    def draw_postprocessing(self, cid):
        self.elements[cid].get_element().update()

        elements_info = {}
        dependency = []

        for key in self.elements.keys():
            w = self.elements[key]
            if w.get_id() == "root":
                continue

            elements_info[w.get_id()] = {"parent": w.get_parent(), "type": w.get_id()}
            dependency.append((w.get_id(), w.get_parent()))

        digraph = nx.DiGraph(dependency)
        order = list((list(nx.topological_sort(digraph))))

        for element_key in order:
            if element_key == "root":
                continue

            cur_element = self.elements[element_key]

            attributes = cur_element.get_attributes_list()

            if (
                AttributeNames.height in attributes
                and AttributeNames.width in attributes
            ):
                self.postprocessing_size_setter(cid, cur_element, attributes)

    def postprocessing_size_setter(self, cid, cur_element, attributes):
        """
        Postprocessing - Used to enable setting the size and width of tkinter elemnents independent.
        Normal behavior of tkinter: Only set set the size the user wants when setting both height and width.
        Wished behavior: Also do this if only one of height or width is set.

        This is implemented here, by first analyzing which elements are drawn latest.
        Then from the latest to the first compute their actual sizes, and if necessary correct them.
        In this manner every element takes the size it has been assigned.
        """
        true_width = cur_element.get_element().winfo_width()
        true_height = cur_element.get_element().winfo_height()

        config_width = int(
            cur_element.get_attributes_list()[AttributeNames.width]["value"]
        )
        config_height = int(
            cur_element.get_attributes_list()[AttributeNames.height]["value"]
        )

        pack = False

        if config_width > 0:
            pack = True
            cur_element.get_element().config(width=config_width)
        else:
            cur_element.get_element().config(width=true_width)

        if config_height > 0:
            pack = True
            cur_element.get_element().config(height=config_height)
        else:
            cur_element.get_element().config(height=true_height)

        if pack:
            if AttributeNames.child_layout in attributes:
                operation = cur_element.get_attributes_list()[
                    AttributeNames.child_layout
                ]["value"]
                if operation == ChildLayoutType.FLEX:
                    cur_element.get_element().pack_propagate(0)
            else:
                cur_element.get_element().pack_propagate(0)

            self.elements[cid].get_element().update()

    def draw(self, cid):
        self.first = False

        self.elements[cid].get_element().mainloop()



"""
Module that takes care of the custom logging config, so it contains the Logger-class.
"""

import logging
import os

BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE = range(8)

COLOR_NAMES = {
    "YELLOW": YELLOW,
    "GREEN": GREEN,
    "BLUE": BLUE,
    "RED": RED,
    "MAGENTA": MAGENTA,
    "CYAN": CYAN,
}

RESET_SEQ = "\033[0m"
COLOR_SEQ = "\033[1;%dm"
BOLD_SEQ = "\033[1m"


def colored_text(txt, color):
    """
    Adds a color to the given text
    """
    return COLOR_SEQ % (30 + COLOR_NAMES[color]) + txt + RESET_SEQ


def uictl_log(text):
    """
    Adds the color of the ui control to the given text for logging
    """
    return colored_text(text, "CYAN")


def domctl_log(text):
    """
    Adds the color of the domain control to the given text for logging
    """
    return colored_text(text, "MAGENTA")


class Logger:
    """
    Provides methods to set the logging config appropriatly. In principle two loggers exists -
    one for the client (default name: clinguin_client) and one for the server (default name: clinguin_server).
    """

    class ColoredFormatter(logging.Formatter):
        """
        For colored logs.
        """

        def __init__(self, msg, use_color=True):
            logging.Formatter.__init__(self, msg)
            self.use_color = use_color

        def format(self, record):
            levelname = record.levelname
            if self.use_color and levelname in Logger.COLORS:
                levelname_color = colored_text(levelname[0:4], Logger.COLORS[levelname])
                record.levelname = levelname_color
            return logging.Formatter.format(self, record)

    client_logger_name = "clinguin_client"
    server_logger_name = "clinguin_server"

    log_levels = {
        "NOTSET": logging.NOTSET,
        "DEBUG": logging.DEBUG,
        "INFO": logging.INFO,
        "WARNING": logging.WARNING,
        "ERROR": logging.ERROR,
        "CRITICAL": logging.CRITICAL,
    }

    @classmethod
    def formatter_message(cls, message, use_color=True):
        """
        Formats certain aspects of the messages (color).
        """
        if use_color:
            message = message.replace("$RESET", RESET_SEQ).replace("$BOLD", BOLD_SEQ)
        else:
            message = message.replace("$RESET", "").replace("$BOLD", "")
        return message

    COLORS = {
        "WARNING": "YELLOW",
        "INFO": "GREEN",
        "DEBUG": "BLUE",
        "CRITICAL": "RED",
        "ERROR": "RED",
    }

    @classmethod
    def _get_log_file_path(cls, log_arg_dict):
        log_file_path = os.path.join(
            "logs", (log_arg_dict["timestamp"] + "-" + log_arg_dict["name"] + ".log")
        )
        return log_file_path

    @classmethod
    def _add_shell_handler_to_logger(cls, logger, log_arg_dict):
        shell_formatter = Logger.ColoredFormatter(log_arg_dict["format_shell"])

        logger.setLevel(cls.log_levels[log_arg_dict["level"]])

        handler_sh = logging.StreamHandler()
        handler_sh.setFormatter(shell_formatter)

        logger.addHandler(handler_sh)

    @classmethod
    def _add_file_handler_to_logger(cls, logger, log_arg_dict, log_file_path):
        file_formatter = Logger.ColoredFormatter(log_arg_dict["format_file"])

        with open(log_file_path, "a+", encoding="utf-8") as file_object:
            file_object.write(
                "<<<<<NEW-LOG-INSTANCE-" + log_arg_dict["name"] + ">>>>>\n\n"
            )

        handler_f = logging.FileHandler(log_file_path)
        handler_f.setFormatter(file_formatter)
        logger.addHandler(handler_f)

    @classmethod
    def setup_logger(cls, log_arg_dict, process=None):
        """
        Function for setting up the logger.
        """
        if process and process == "client":
            cls.client_logger_name = log_arg_dict["name"]
        elif process and process == "server":
            cls.server_logger_name = log_arg_dict["name"]

        log_file_path = cls._get_log_file_path(log_arg_dict)

        logger = logging.getLogger(log_arg_dict["name"])
        if not log_arg_dict["shell_disabled"]:
            cls._add_shell_handler_to_logger(logger, log_arg_dict)
        if log_arg_dict["file_enabled"]:
            cls._add_file_handler_to_logger(logger, log_arg_dict, log_file_path)

    @classmethod
    def setup_uvicorn_logger_on_startup(cls, log_arg_dict):
        """
        General function for setting up the logger on startup.
        """
        logger = logging.getLogger("uvicorn.access")
        for handler in logger.handlers:
            logger.removeHandler(handler)

        logger = logging.getLogger("uvicorn.error")
        for handler in logger.handlers:
            logger.removeHandler(handler)

        logger = logging.getLogger("uvicorn")
        for handler in logger.handlers:
            logger.removeHandler(handler)

        # Add new handlers to top-level-uvicorn logger
        log_file_path = cls._get_log_file_path(log_arg_dict)

        logger = logging.getLogger("uvicorn")
        if not log_arg_dict["shell_disabled"]:
            cls._add_shell_handler_to_logger(logger, log_arg_dict)
        if log_arg_dict["file_enabled"]:
            cls._add_file_handler_to_logger(logger, log_arg_dict, log_file_path)



# pylint: disable=E1101
"""
This module contains the ConfigureBorder class.
"""

from clinguin.utils.attribute_types import ColorType, IntegerType

from .attribute_names import AttributeNames
from .extension_class import ExtensionClass


class ConfigureBorder(ExtensionClass):
    """
    If a element shall have a border (e.g. container), it must be a subtype of this class.
    """

    @classmethod
    def get_attributes(cls, attributes=None):
        if attributes is None:
            attributes = {}

        attributes[AttributeNames.border_width] = {
            "value": 0,
            "value_type": IntegerType,
        }
        attributes[AttributeNames.border_color] = {
            "value": "black",
            "value_type": ColorType,
        }

        return attributes

    def _set_border_width(
        self, elements, key=AttributeNames.border_width
    ):  # pylint: disable=W0613
        value = self._attributes[key]["value"]
        if value > 0:
            # Not using borderwidth as one cannot set the color of the default border
            self._element.configure(highlightthickness=int(value))
        elif value == 0:
            # Zero is perfectly fine, but it shouldn't be configured then
            pass
        else:
            self._logger.warning(
                "For element "
                + self._id
                + " ,setBorderwidth for "
                + key
                + " is lesser than 0: "
                + str(value)
            )

    def _set_border_background_color(
        self, elements, key=AttributeNames.border_color
    ):  # pylint: disable=W0613
        # Not using borderwidth as one cannot set the color of the default border
        value = self._attributes[key]["value"]
        self._element.configure(highlightbackground=value, highlightcolor=value)



# pylint: disable=R0903
"""
Module that contains the AttributeDao.
"""
from clorm import Predicate, RawField


class AttributeDao(Predicate):
    """
    Attributes define the style of the UI.

    - id: Identifier of the element that the attribute will be set to.
    - key: The name of the attribute. Available attributes depend on the element type and the frontend.
    - value: The value of the attribute.
    """

    # pylint: disable=abstract-method

    id = RawField
    key = RawField
    value = RawField

    class Meta:
        """
        Meta class
        """

        name = "attr"



# pylint: disable=E1101,R0801
"""
This module contains the ConfigureSize class.
"""

from clinguin.utils.attribute_types import ChildLayoutType, IntegerType

from .attribute_names import AttributeNames
from .extension_class import ExtensionClass


class ConfigureSize(ExtensionClass):
    """
    If a element is a subtype of the configure-size class the size of the element can be adjusted.
    """

    @classmethod
    def get_attributes(cls, attributes=None):
        if attributes is None:
            attributes = {}

        attributes[AttributeNames.height] = {"value": 0, "value_type": IntegerType}
        attributes[AttributeNames.width] = {"value": 0, "value_type": IntegerType}

        return attributes

    def _set_size(self, elements):  # pylint: disable=W0613
        height = self._attributes[AttributeNames.height]["value"]
        width = self._attributes[AttributeNames.width]["value"]

        if height > 0 and width > 0:
            if AttributeNames.child_layout in self._attributes:
                child_layout_value = self._attributes[AttributeNames.child_layout][
                    "value"
                ]

                if child_layout_value in (
                    ChildLayoutType.FLEX,
                    ChildLayoutType.RELSTATIC,
                    ChildLayoutType.ABSSTATIC,
                ):
                    self._element.pack_propagate(0)
                elif child_layout_value == ChildLayoutType.GRID:
                    self._element.grid_propagate(0)
            else:
                self._element.pack_propagate(0)

        if height > 0:
            self._element.configure(height=int(height))

        if width > 0:
            self._element.configure(width=int(width))

        if height < 0:
            self._logger.warning(
                "Height of " + self._id + " has illegal value (" + str(height) + ")"
            )

        if width < 0:
            self._logger.warning(
                "Width of " + self._id + " has illegal value (" + str(height) + ")"
            )



# pylint: disable=E1101
"""
This module contains the LayoutFollower class.
"""
import logging

from clinguin.utils import Logger
from clinguin.utils.attribute_types import (
    ChildLayoutType,
    FlexDirectionType,
    IntegerType,
)

from .attribute_names import AttributeNames
from .extension_class import ExtensionClass


class LayoutFollower(ExtensionClass):
    """
    If a element is a subtype of this class then one can be sure, that one can use it in layouts.
    """

    def __init__(self):
        self._logger = logging.getLogger(Logger.client_logger_name)

    @classmethod
    def get_attributes(cls, attributes=None):
        if attributes is None:
            attributes = {}

        # Layout-Control
        attributes[AttributeNames.flex_direction] = {
            "value": FlexDirectionType.COLUMN,
            "value_type": FlexDirectionType,
        }

        attributes[AttributeNames.grid_column] = {"value": 0, "value_type": IntegerType}
        attributes[AttributeNames.grid_row] = {"value": 0, "value_type": IntegerType}
        attributes[AttributeNames.grid_column_span] = {
            "value": 1,
            "value_type": IntegerType,
        }
        attributes[AttributeNames.grid_row_span] = {
            "value": 1,
            "value_type": IntegerType,
        }

        attributes[AttributeNames.pos_x] = {"value": 0, "value_type": IntegerType}
        attributes[AttributeNames.pos_y] = {"value": 0, "value_type": IntegerType}

        return attributes

    def _set_layout(self, elements):
        # pylint: disable=R0912
        parent = elements[self._parent]
        if hasattr(parent, "get_child_org"):
            parent_org = getattr(parent, "get_child_org")()
        else:
            self._logger.warning(
                "Could not find necessary attribute childOrg() in id: %s",
                str(self._parent),
            )
            return

        if parent_org == ChildLayoutType.FLEX:
            flex_direction_type = self._attributes[AttributeNames.flex_direction][
                "value"
            ]

            flex_direction_tkinter_type = ""
            if flex_direction_type == FlexDirectionType.COLUMN:
                flex_direction_tkinter_type = "top"
            elif flex_direction_type == FlexDirectionType.COLUMN_REVERSE:
                flex_direction_tkinter_type = "bottom"
            elif flex_direction_type == FlexDirectionType.ROW:
                flex_direction_tkinter_type = "left"
            elif flex_direction_type == FlexDirectionType.ROW_REVERSE:
                flex_direction_tkinter_type = "right"

            self._element.pack(
                expand=True, fill="both", side=flex_direction_tkinter_type
            )

        elif parent_org == ChildLayoutType.GRID:
            grid_pos_column = self._attributes[AttributeNames.grid_column]["value"]
            grid_pos_row = self._attributes[AttributeNames.grid_row]["value"]

            grid_span_column = self._attributes[AttributeNames.grid_column_span][
                "value"
            ]
            grid_span_row = self._attributes[AttributeNames.grid_row_span]["value"]

            if (
                int(grid_pos_column) >= 0
                and int(grid_pos_row) >= 0
                and int(grid_span_column) >= 1
                and int(grid_span_row) >= 1
            ):
                self._element.grid(
                    column=grid_pos_column,
                    row=grid_pos_row,
                    columnspan=int(grid_span_column),
                    rowspan=int(grid_span_row),
                )
            else:
                self._logger.warning(
                    "Could not set grid-layout due to illegal values for element: %s",
                    str(self._id),
                )

        elif parent_org in (ChildLayoutType.ABSSTATIC, ChildLayoutType.RELSTATIC):
            self._element.pack(expand=True, fill="both")

            x = self._attributes[AttributeNames.pos_x]["value"]
            y = self._attributes[AttributeNames.pos_y]["value"]

            if int(x) >= 0 and int(y) >= 0:
                if parent_org == ChildLayoutType.ABSSTATIC:
                    self._element.place(x=int(x), y=int(y))
                elif parent_org == ChildLayoutType.RELSTATIC:
                    self._element.place(relx=int(x) / 100, rely=int(y) / 100)
                else:
                    error_string = (
                        "For element "
                        + self._id
                        + " ,either posx or posy are not non-negative-numbers. "
                    )
                    self._logger.error(error_string)
                    raise Exception(error_string)
            else:
                self._logger.warning(
                    "For element %s invalid positioning supplied.", self._id
                )
        else:
            self._logger.warning(
                "For element %s no child layout was defined.", self._id
            )



"""
This module contains the IntegerType class.
"""

from .type import Type
from .utils.standard_text_processing import StandardTextProcessing


class IntegerType(Type):
    """
    The IntegerType shall be used, when the attribute value is integer like.
    """

    @classmethod
    def parse(cls, parse_input: str, logger):
        parsed_string = StandardTextProcessing.parse_string_with_quotes(parse_input)

        try:
            return int(parsed_string)
        except Exception as ex:
            logger.error("Could not parse string to int: " + parsed_string)
            raise Exception("Could not parse string to int: " + parsed_string) from ex

    @classmethod
    def description(cls):
        return "Specification as an integer, can be either specified as a symbol or a string."



"""
Responsible for starting the server
"""

import uvicorn
from fastapi import FastAPI
from starlette.middleware.cors import CORSMiddleware

from clinguin.server import Endpoints
from clinguin.utils import Logger


def start(args):
    """
    Function that starts the uvicorn server.
    """

    app = FastAPI(description="clinguin")

    # Disable CORS
    app.add_middleware(
        CORSMiddleware,
        allow_origins=["*"],
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )

    @app.on_event("startup")
    async def startup_event():
        Logger.setup_uvicorn_logger_on_startup(args.log_args)

    endpoints = Endpoints(args)
    app.include_router(endpoints.router)

    uvicorn.run(app, port=args.server_port)



"""
Module which contains all possible attribute- and callback-types.
"""

from .boolean import BooleanType
from .child_layout import ChildLayoutType
from .color import ColorType
from .enum import EnumType
from .flex_direction import FlexDirectionType
from .float import FloatType
from .font_families import FontFamiliesType
from .font_weight import FontWeightType
from .image import ImageType
from .integer import IntegerType
from .path import PathType
from .popup_types import PopupTypesType
from .string import StringType
from .symbol import SymbolType
from .type import Type

__all__ = [
    Type.__name__,
    BooleanType.__name__,
    IntegerType.__name__,
    FloatType.__name__,
    StringType.__name__,
    ColorType.__name__,
    EnumType.__name__,
    ChildLayoutType.__name__,
    SymbolType.__name__,
    ImageType.__name__,
    PathType.__name__,
    PopupTypesType.__name__,
    FontFamiliesType.__name__,
    FontWeightType.__name__,
    FlexDirectionType.__name__,
]



"""
Module that contains the FlexDirectionType
"""

from enum import auto

from .enum import EnumType
from .utils.standard_text_processing import StandardTextProcessing


class FlexDirectionType(EnumType):
    """
    The FlexDirectionType is an Enum Type, which can be used to specify where a component shall
    be added relative to the parent.
    """

    COLUMN = auto()
    COLUMN_REVERSE = auto()
    ROW = auto()
    ROW_REVERSE = auto()

    @classmethod
    def parse(cls, parse_input: str, logger):
        parsed_string = (
            StandardTextProcessing.parse_string_with_quotes(parse_input)
        ).lower()

        return_value = None

        if parsed_string == cls.COLUMN.name.lower():
            return_value = cls.COLUMN
        elif parsed_string == cls.COLUMN_REVERSE.name.lower():
            return_value = cls.COLUMN_REVERSE
        elif parsed_string == cls.ROW.name.lower():
            return_value = cls.ROW
        elif parsed_string == cls.ROW_REVERSE.name.lower():
            return_value = cls.ROW_REVERSE
        else:
            logger.error(
                "Could not parse " + parsed_string + " to flex_direction type."
            )
            raise Exception(
                "Could not parse " + parsed_string + " to flex_direction type."
            )

        return return_value

    @classmethod
    def description(cls):
        return (
            "For the flex-direction type two possible values exist: column (vertical alignment)"
            + "and row (horizontal alignment)."
        )



"""
This module contains the SymbolType class.
"""

import clingo

from .type import Type
from .utils.standard_text_processing import StandardTextProcessing


class SymbolType(Type):
    """
    This class checks if the input IS A symbol, it does not parse it to a symbol.
    """

    @classmethod
    def parse(cls, parse_input: str, logger) -> str:
        parsed_string = StandardTextProcessing.parse_string_with_quotes(parse_input)

        try:
            clingo.parse_term(parsed_string)
            return parsed_string
        except Exception as ex:
            error_string = "The string " + parsed_string + " is not a clingo symbol!"
            logger.error(error_string)
            raise Exception(error_string) from ex

    @classmethod
    def description(cls):
        return "Corresponds to a clingo-symbol, e.g. p(4)"



"""
Contains the canvas class.
"""

import base64
import io
import tkinter as tk

import PIL.Image as TImage
from PIL import ImageTk

from clinguin.utils.attribute_types import BooleanType, ImageType, PathType

from ..tkinter_utils import AttributeNames, ConfigureSize, LayoutFollower
from .root_cmp import RootCmp


class Canvas(RootCmp, LayoutFollower, ConfigureSize):
    """
    The canvas class resembles a element,
    which is generally regarded as an ''empty sheet of paper'',
    so there is an area, where one can display images, draw something or does something similar.
    One can look up what actual options are available through the syntax-definition.
    Implementation wise it is similarly implemented to the label, button and dropdownmenu.
    """

    def __init__(self, args, cid, parent, attributes, callbacks, base_engine):
        super().__init__(args, cid, parent, attributes, callbacks, base_engine)

        self._canvas = None
        self._image = None

    def _init_element(self, elements):
        canvas_frame = tk.Frame(elements[str(self._parent)].get_element())

        self._canvas = tk.Canvas(canvas_frame)

        return canvas_frame

    @classmethod
    def _get_attributes(cls, attributes=None):
        if attributes is None:
            attributes = {}

        attributes[AttributeNames.image] = {"value": "", "value_type": ImageType}
        attributes[AttributeNames.image_path] = {"value": "", "value_type": PathType}
        attributes[AttributeNames.resize] = {"value": True, "value_type": BooleanType}
        return attributes

    def _set_image(self, elements):
        self._logger.debug(str(elements))

        image_base64 = self._attributes[AttributeNames.image]["value"]
        image_file = self._attributes[AttributeNames.image_path]["value"]

        if image_base64 != "" and image_file != "":
            self._logger.error(
                "One cannot set both attributes %s and %s for the same canvas with id %s",
                AttributeNames.image,
                AttributeNames.image_path,
                str(self._id),
            )

        elif image_base64 == "" and image_file != "":
            try:
                image_open = TImage.open(image_file)

                height = self._attributes[AttributeNames.height]["value"]
                width = self._attributes[AttributeNames.width]["value"]
                resize_flag = self._attributes[AttributeNames.resize]["value"]

                if height > 0 and width > 0 and resize_flag:
                    image_open = image_open.resize((width, height))

                tkinter_image = ImageTk.PhotoImage(image_open, master=self._canvas)

                self._canvas.create_image(0, 0, anchor=tk.NW, image=tkinter_image)

                # DO NOT DELETE THIS SEEMINGLY UNECESSARY LINE (this is due to a tkinter bug)
                self._image = tkinter_image
            except Exception as e:
                self._logger.error(e)
                self._logger.error(
                    "Could not render image (likely the provided file <%s>is not a valid image).",
                    image_file,
                )

        elif image_base64 != "" and image_file == "":
            try:
                image_initial_bytes = image_base64.encode("utf-8")
                image_decoded = base64.b64decode(image_initial_bytes)

                image_bytes = io.BytesIO(image_decoded)
                image_open = TImage.open(image_bytes)

                height = self._attributes[AttributeNames.height]["value"]
                width = self._attributes[AttributeNames.width]["value"]
                resize_flag = self._attributes[AttributeNames.resize]["value"]

                if height > 0 and width > 0 and resize_flag:
                    image_open = image_open.resize((width, height))

                tkinter_image = ImageTk.PhotoImage(image_open, master=self._canvas)

                self._canvas.create_image(0, 0, anchor=tk.NW, image=tkinter_image)

                # DO NOT DELETE THIS SEEMINGLY UNECESSARY LINE
                self._image = tkinter_image
            except Exception:
                self._logger.error(
                    "Could not render image (likely Base64 encoding is wrong)."
                )

    def _add_component_to_elements(self, elements):
        self._canvas.pack(expand=True, fill="both")

        elements[str(self._id)] = self



# pylint: disable=E1101
"""
This module contains the LayoutController class.
"""

from clinguin.utils.attribute_types import ChildLayoutType

from .attribute_names import AttributeNames
from .extension_class import ExtensionClass


class LayoutController(ExtensionClass):
    """
    If a element is a subtype of this class, then it can steer the layouting behavior of children.
    """

    def __init__(self):
        self._child_layout = None

    @classmethod
    def get_attributes(cls, attributes=None):
        if attributes is None:
            attributes = {}

        attributes[AttributeNames.child_layout] = {
            "value": ChildLayoutType.FLEX,
            "value_type": ChildLayoutType,
        }

        return attributes

    def _set_child_org(self, elements):  # pylint: disable=W0613
        value = self._attributes[AttributeNames.child_layout]["value"]
        self._child_layout = value

    def get_child_org(self):
        """
        Returns which child layout is used.
        """
        return self._child_layout

    def get_fit_children_size(self):
        """
        Returns the variable associated with ''if fit to children size''.
        """
        return self._fit_children_size



"""
Module contains the AttributeDto class.
"""

import json


class AttributeDto:
    """
    Objects of this class represent the attributes in the json convertible hierarchy.
    """

    def __init__(self, cid, key, value):
        self.id = str(cid)  # pylint: disable=C0103
        self.key = str(key)
        self.value = str(value)

    def to_JSON(self):  # pylint: disable=C0103
        """
        Converts the object (self) to a json representation.
        """
        return json.dumps(self, default=lambda o: o.__dict__)

    def clone(self):
        """
        Creates a new AttributeDto object with the same properties.
        """
        return AttributeDto(self.id, self.key, self.value)



"""
Module for the Endpoints class.
"""

import logging
import traceback
from importlib.metadata import metadata

import clingo
from fastapi import APIRouter

from clinguin.utils import Logger

from ...utils import get_server_error_alert
from ...utils.logger import colored_text
from .backend_operation_dto import BackendOperationDto

# Self Defined
from .endpoints_helper import EndpointsHelper


class Endpoints:
    """
    The endpoints class define the available endpoints the backend
    (this time backend refers to the general concept of backend, like a server backend) has.
    These are defined in the ''__init__'', and correspond to the three methods:

    Methods:
        health -> Json : Returns name, version and description of clinguin.
        standard_executor -> Json : Returns the default GUI representation as Json that the Backend provides.
        operation_executor -> Json : Executes a operation defined by the Json passed with the Post request.
    """

    last_response = None

    def __init__(self, args) -> None:
        Logger.setup_logger(args.log_args, process="server")
        self._logger = logging.getLogger(args.log_args["name"])

        self.router = APIRouter()

        # Definition of endpoints
        self.router.add_api_route("/health", self.health, methods=["GET"])
        self.router.add_api_route("/", self.standard_executor, methods=["GET"])
        self.router.add_api_route("/backend", self.operation_executor, methods=["POST"])

        self._backend = args.backend(args)

    async def health(self):
        """
        Health endpoint (/health) of the server-backend,
        returns name, version and summary of clinguin.
        """
        self._logger.info("--> Health")
        cuin = metadata("clinguin")
        return {
            "name": cuin["name"],
            "version": cuin["version"],
            "description": cuin["summary"],
        }

    async def standard_executor(self):
        """
        Get endpoint (/) of the server-backend,
        calls the get() method of the respective backend (clinguin/clingo/clingraph/etc.).
        The get() method is implemented by every backend.
        """
        self._logger.info(colored_text("=>=>=>=>=>=>=>=>=> GET", "GREEN"))
        self._logger.info("%s get()", colored_text("-->", "GREEN"))
        try:
            json = self._backend.get()
            self.last_response = json
            self._logger.info(colored_text("--------------------", "GREEN"))
            return json
        except Exception as e:
            self._logger.error("Handling global exception in endpoint")
            self._logger.error(e)
            self._logger.error(traceback.format_exc())
            return get_server_error_alert(str(e), self.last_response)

    async def operation_executor(self, backend_call_string: BackendOperationDto):
        """
        Post endpoint (/backend) of the server-backend,
        which can be used to call specific methods/functions of the backend.
        It takes a backend_call_string, which is essentially a json defining a function to be called,
        including arguments.
        For example: {'function':'add_assumption(p(1),true)'}
        """

        self._logger.info(colored_text("=>=>=>=>=>=>=>=>=> POST", "GREEN"))

        try:
            try:
                symbol = clingo.parse_term(backend_call_string.function)
            except Exception as exc:
                msg = f"Could not parse {backend_call_string.function} into an atom."
                self._logger.error(msg)
                raise Exception(msg) from exc

            if symbol.type != clingo.SymbolType.Function:
                raise Exception(f"Operation {symbol} is not a function")

            # pylint: disable=protected-access
            if hasattr(backend_call_string, "context"):
                self._backend._set_context(backend_call_string.context)
            else:
                self._backend._set_context([])

            function_name = symbol.name
            operations = []
            if function_name == "":
                operations = symbol.arguments
            else:
                operations = [symbol]

            for p in operations:
                function_name = p.name

                function_arguments = list(map(str, p.arguments))

                call_args = ",".join(function_arguments)
                self._logger.info(
                    "%s %s(%s)", colored_text("-->", "GREEN"), function_name, call_args
                )

                EndpointsHelper.call_function(
                    self._backend, function_name, function_arguments, {}
                )

            self._logger.info(" get(%s)", colored_text("-->", "GREEN"))

            self.last_response = self._backend.get()
            self._logger.info(colored_text("--------------------", "GREEN"))
            return self.last_response

        except Exception as e:
            self._logger.error("Handling global exception in endpoint")
            self._logger.error(e)
            self._logger.error(traceback.format_exc())
            return get_server_error_alert(str(e), self.last_response)



"""
Module that contains the ClingoDL Backend.
"""

from clingodl import ClingoDLTheory
from clinguin.server.application.backends.theory_backend import TheoryBackend


class ClingoDLBackend(TheoryBackend):
    """
    Backend that allows programs using clingodl theory atoms as input.
    It also includes the assignment in the domain state.
    """

    theory_class = ClingoDLTheory



"""
Module that contains the FontWeightType.
"""

import itertools
from enum import auto

from .enum import EnumType
from .utils.standard_text_processing import StandardTextProcessing


class FontWeightType(EnumType):
    """
    The FontWeightType is an Enum Type, which can be used to specify look and feel of text.
    """

    NORMAL = auto()
    ITALIC = auto()
    BOLD = auto()
    ITALIC_BOLD = auto()

    @classmethod
    def parse(cls, parse_input: str, logger):
        parsed_string = (
            StandardTextProcessing.parse_string_with_quotes(parse_input)
        ).lower()

        normals = ["n", "normal"]
        bolds = ["b", "bold"]
        italics = ["i", "italic"]

        combination = [bolds, italics]

        bolds_italic_list = []
        for p in itertools.permutations(combination, 2):
            l0 = p[0]
            l1 = p[1]
            zipped = list(zip(l0, l1))
            for l in zipped:
                bolds_italic_list.append(str(l[0]) + str(l[1]))
                bolds_italic_list.append(str(l[0]) + "_" + str(l[1]))
                bolds_italic_list.append(str(l[0]) + "-" + str(l[1]))

        return_value = None

        if parsed_string in bolds_italic_list:
            return_value = cls.ITALIC_BOLD
        elif parsed_string in bolds:
            return_value = cls.BOLD
        elif parsed_string in italics:
            return_value = cls.ITALIC
        elif parsed_string in normals:
            return_value = cls.NORMAL
        else:
            error_string = "Could not parse " + parsed_string + " to FontWeightType."
            logger.error(error_string)
            raise Exception(error_string)

        return return_value

    @classmethod
    def description(cls):
        return (
            "For the FontWeightType several different options exists - ['i' or 'italic'] for italic,"
            + "['b' or 'bold'] for bold and any permuation (with the delimiters '_', '-' and no delimiter)"
            + "of those for both italic and bolt. Further 'normal' exists for specifying normal font weight."
        )



"""
This module provides different errors for clinguin.
"""


class NoModelError(Exception):
    """
    The NoModelError is raised, if the provided logic program is unsatisfiable.
    """

    def __init__(self, core=None) -> None:
        super().__init__()
        self.core = core


def get_server_error_alert(message="", last_response=None):
    """
    Returns a JSON, which corresponds to valid clinguin json syntax and displays an error message.
    """
    error_alert = {
        "id": "server_error",
        "type": "message",
        "parent": "window",
        "attributes": [
            {
                "id": "server_error",
                "key": "message",
                "value": message + " (Check the server logs for more details).",
            },
            {"id": "server_error", "key": "title", "value": "Server error"},
            {"id": "server_error", "key": "type", "value": "danger"},
        ],
        "when": [],
        "children": [],
    }

    if last_response is None:
        return {
            "ui": {
                "id": "root",
                "type": "root",
                "parent": "root",
                "attributes": [],
                "when": [],
                "children": [
                    {
                        "id": "window",
                        "type": "window",
                        "parent": "root",
                        "attributes": [],
                        "when": [],
                        "children": [error_alert],
                    }
                ],
            },
            "ds": {},
        }

    last_response["ui"].children[0].children.append(error_alert)
    return last_response



from reference_json_output import (
    BasicTest06,
    BasicTest07,
    BasicTest08,
    BasicTest09,
    BasicTest10,
    BasicTest11,
    BasicTest12,
)
from utils_test_utils import UtilsTestUtils


class TestBackendUnit_06_11:
    def setup_method(self, test_method):
        self.backend = UtilsTestUtils.instantiate_backend(test_method)

    def teardown_method(self, test_method):
        pass

    def test_basic_12(self):
        should_output = BasicTest12.get_reference_json()
        received_by_backend = self.backend.get()
        UtilsTestUtils.assert_result(should_output, received_by_backend["ui"])

    def test_basic_11(self):
        should_output = BasicTest11.get_reference_json()
        received_by_backend = self.backend.get()
        UtilsTestUtils.assert_result(should_output, received_by_backend["ui"])

    def test_basic_10(self):
        should_output = BasicTest10.get_reference_json()
        received_by_backend = self.backend.get()
        UtilsTestUtils.assert_result(should_output, received_by_backend["ui"])

    def test_basic_09_order_02(self):
        should_output = BasicTest09.get_reference_json()
        received_by_backend = self.backend.get()
        UtilsTestUtils.assert_result(should_output, received_by_backend["ui"])

        self.backend.add_assumption("p(2)", "true")
        received_by_backend = self.backend.get()
        should_output = BasicTest09.post_p_1_reference_json()
        UtilsTestUtils.assert_result(should_output, received_by_backend["ui"])

        should_output = BasicTest09.get_p_2_reference_json()
        received_by_backend = self.backend.get()
        UtilsTestUtils.assert_result(should_output, received_by_backend["ui"])

        should_output = BasicTest09.post_p_3_reference_json()
        self.backend.clear_assumptions()
        received_by_backend = self.backend.get()
        UtilsTestUtils.assert_result(should_output, received_by_backend["ui"])

        should_output = BasicTest09.get_reference_json()
        received_by_backend = self.backend.get()
        UtilsTestUtils.assert_result(should_output, received_by_backend["ui"])

    def test_basic_09_order_01(self):
        should_output = BasicTest09.get_reference_json()
        received_by_backend = self.backend.get()
        UtilsTestUtils.assert_result(should_output, received_by_backend["ui"])

    def test_basic_08(self):
        should_output = BasicTest08.get_reference_json()
        received_by_backend = self.backend.get()
        UtilsTestUtils.assert_result(should_output, received_by_backend["ui"])

    def test_basic_07(self):
        should_output = BasicTest07.get_reference_json()
        received_by_backend = self.backend.get()
        UtilsTestUtils.assert_result(should_output, received_by_backend["ui"])

    def test_basic_06(self):
        should_output = BasicTest06.get_reference_json()
        received_by_backend = self.backend.get()
        UtilsTestUtils.assert_result(should_output, received_by_backend["ui"])



"""
This module contains the FloatType class.
"""

from .type import Type
from .utils.standard_text_processing import StandardTextProcessing


class FloatType(Type):
    """
    The flaot type shall be used, when the attribute-value is float like.
    """

    @classmethod
    def parse(cls, parse_input: str, logger):
        parsed_string = StandardTextProcessing.parse_string_with_quotes(parse_input)

        try:
            return float(parsed_string)
        except Exception as ex:
            logger.error("Could not parse string to float: " + parsed_string)
            raise Exception("Could not parse string to float: " + parsed_string) from ex

    @classmethod
    def description(cls):
        return 'Is a float type. A float must be specified as a string (e.g. "5.98").'



"""
Module that contains the Clingcon Backend.
"""

from clingcon import ClingconTheory
from clinguin.server.application.backends.theory_backend import TheoryBackend


class ClingconBackend(TheoryBackend):
    """
    Backend that allows programs using clingcon theory atoms as input.
    It also includes the assignment in the domain state.
    """

    theory_class = ClingconTheory



"""
Clingo logger.
TODO
"""

from clingo import MessageCode


class ClingoLogger:
    """
    ClingoLogger Class.
    TODO
    """

    errors: list[str] = []

    @classmethod
    def logger(cls, msgC: MessageCode, errStr: str) -> None:
        """
        Adds a log.
        """
        print(msgC, " : ", errStr)
        cls.errors.append(errStr)

    @classmethod
    def errorString(cls):
        """
        Adds an error string.
        """
        return "\n".join(cls.errors)



"""
This module contains the CallbackNames class.
"""


class CallbackNames:  # pylint: disable=R0903
    """
    This class contains all possible names for actions for callbacks and their descriptions.
    """

    click = "click"
    clear = "clear"

    descriptions = {
        click: "The click action, is the action that is executed if one clicks with the left-mouse button on a"
        + "element.",
        clear: "The clearlick action, is the action that is executed when the empty option of a dropdown menu is"
        + "selected.",
    }



# pylint: disable=R0801
"""
Contains the menu bar section item class.
"""

from clinguin.utils.attribute_types import StringType, SymbolType

from ..tkinter_utils import AttributeNames, CallBackDefinition, CallbackNames
from .root_cmp import RootCmp

key_map = {
    "Ctrl": "Control",
    "Opt": "Option",
    "Cmd": "Command",
    "Control": "Control",
    "Option": "Option",
    "Command": "Command",
    "Shift": "Shift",
}


def accelerator_to_bind(a):
    """
    Getting key bindings as string.
    """
    args = a.split("+")
    formatted = []
    for e in args:
        if e in key_map:
            formatted.append(key_map[e])
        else:
            formatted.append(e.lower())
    return "<" + "-".join(formatted) + ">"


class MenuBarSectionItem(RootCmp):
    """
    The menu bar section is a section of a menu bar (e.g. in the menu |main|contact|,
    where if one clicks on |contact| further the options |location|team| appear,
    a menu-bar-section would be |contact|, whereas |location| and |team| would be menu-bar-section-items.
    """

    def _init_element(self, elements):
        menubar_section = elements[self._parent].get_element()

        return menubar_section

    @classmethod
    def _get_attributes(cls, attributes=None):
        if attributes is None:
            attributes = {}

        attributes[AttributeNames.label] = {
            "value": "standard_label",
            "value_type": StringType,
        }
        attributes[AttributeNames.accelerator] = {
            "value": None,
            "value_type": StringType,
        }

        return attributes

    @classmethod
    def _get_callbacks(cls, callbacks=None):
        if callbacks is None:
            callbacks = {}

        callbacks[CallbackNames.click] = {
            "operation": None,
            "operation_type": SymbolType,
        }

        return callbacks

    def _define_click_event(self, elements):
        key = CallbackNames.click
        text = self._attributes[AttributeNames.label]["value"]
        accelerator = self._attributes[AttributeNames.accelerator]["value"]
        if self._callbacks[key] and self._callbacks[key]["operation"]:
            cb = CallBackDefinition(
                self._id,
                self._parent,
                self._callbacks[key]["operation"],
                elements,
                self._menubar_item_click,
            )
            self._element.add_command(label=text, command=cb, accelerator=accelerator)
            menu = elements[self._parent]
            window = elements[menu.get_parent()]
            root = elements[window.get_parent()]
            if accelerator:
                root.get_element().bind(accelerator_to_bind(accelerator), cb)
        else:
            self._element.add_command(label=text)

    def _menubar_item_click(self, cid, parent, click_operation, elements):
        self._logger.debug(str(cid))
        self._logger.debug(str(parent))
        self._logger.debug(str(elements))

        if click_operation is not None:
            self._base_engine.post_with_operation(click_operation)

    def _add_component_to_elements(self, elements):
        elements[str(self._id)] = self

    def forget_children(self, elements):
        self._logger.debug(str(elements))
        # pylint: disable=W0107
        # pylint: enable=W0107



# pylint: disable=R0904
"""
Module that contains the UI State class.
"""
import logging
import os
from pathlib import Path

import clorm
from clingo import Control, parse_term
from clingo.symbol import Function, Number, String
from clorm import Raw

from clinguin.server.data.clinguin_context import ClinguinContext
from clinguin.utils import StandardTextProcessing, image_to_b64

from ...utils.logger import uictl_log
from .attribute import AttributeDao
from .callback import WhenDao
from .element import ElementDao

log = logging.getLogger("clinguin_server")


class UIState:
    """
    The UIState is the low-level-access-class for handling the facts defining the UI state
    """

    unifiers = [ElementDao, AttributeDao, WhenDao]

    def __init__(self, ui_files, domain_state, constants_arg_list):
        self._factbase = None
        self._ui_files = ui_files
        self._domain_state = domain_state
        self._constants_arg_list = constants_arg_list

    def __str__(self):
        s = "\nUI State:\n=========\n"
        s += self._factbase.asp_str()
        return s

    @property
    def is_empty(self):
        """
        Checks whether the factbase is empty.
        if so return true, else false.
        """
        return self._factbase is None

    def _set_fb_symbols(self, symbols):
        self._factbase = clorm.unify(self.unifiers, symbols)

    def ui_control(self):
        """
        Generates a ClingoControl Object to compute the UI state

        """
        log.debug(
            uictl_log(
                f'uictl = Control(["0", "--warn=none"] + {self._constants_arg_list})'
            )
        )
        uictl = Control(["0", "--warn=none"] + self._constants_arg_list)
        for f in self._ui_files:
            path = Path(f)
            if not path.is_file():
                log.critical("File %s does not exist", f)
                raise Exception(f"File {f} does not exist")
            try:
                uictl.load(str(f))
            except Exception as e:
                log.critical(
                    "Failed to load file %s (there is likely a syntax error in this logic program file).",
                    f,
                )
                log.critical(str(e))
                raise e

        log.debug(uictl_log(f'uictl.add("base", [], {self._domain_state})'))
        try:
            uictl.add("base", [], self._domain_state)
        except RuntimeError as e:
            message = """The domain state is not well constructed.\
 Make sure there are no #show statements in the domain files that have tuples as output (without function name).\
 Consider turing on the debug logs with server-log-level=DEBUG to inspect the domain state"""
            log.critical(message)
            raise e
        log.debug(
            uictl_log(
                'uictl.add("base", [], "#show elem/3. #show attr/3. #show when/4.")'
            )
        )
        uictl.add("base", [], "#show elem/3. #show attr/3. #show when/4.")

        log.debug(uictl_log('uictl.ground([("base", [])], ClinguinContext())'))
        uictl.ground([("base", [])], ClinguinContext())
        return uictl

    def update_ui_state(self):
        """
        Computes the answer set representing the UI state
        """
        log.debug("Computing UI state\n")
        uictl = self.ui_control()

        defined = False
        log.debug(uictl_log("uictl.solve(yield_=True)"))
        with uictl.solve(yield_=True) as result:
            for m in result:
                model_symbols = m.symbols(shown=True, atoms=True)
                defined = True
                break
        if not defined:
            log.critical("UI encoding was UNSATISFIABLE")
            raise RuntimeError("UI encoding was UNSATISFIABLE")
        self._factbase = clorm.unify(self.__class__.unifiers, model_symbols)
        # log.debug(uictl_log(". ".join([str(s) for s in model_symbols])))
        # log.debug(uictl_log(self._factbase.asp_str(width=100000, commented=True)))

    def add_message(self, title, message, attribute_type="info"):
        """
        Adds a ''Message'' (aka. Notification/Pop-Up) for the user with a certain title and message.
        """
        windows = list(
            self._factbase.query(ElementDao)
            .where(ElementDao.type == Raw(Function("window", [])))
            .all()
        )
        if len(windows) == 0:
            raise ValueError(
                "No window found to add message. Make sure an element of type window appears in your UI"
            )
        mid = f"{hash(message)}"
        self.add_element(mid, "message", windows[0].symbol.arguments[0])
        self.add_attribute(mid, "title", title)
        self.add_attribute(mid, "message", message)
        self.add_attribute(mid, "type", attribute_type)

    # Manage factbase

    def add_element(self, cid, t, parent):
        """
        Adds an element to the factbase.
        """
        if isinstance(cid, str):
            cid = Function(cid, [])
        if isinstance(t, str):
            t = Function(t, [])
        if isinstance(parent, str):
            parent = Function(parent, [])
        self._factbase.add(ElementDao(Raw(cid), Raw(t), Raw(parent)))

    def add_attribute(self, cid, key, value):
        """
        Adds an attribute to the factbase.
        """
        if isinstance(cid, str):
            cid = Function(cid, [])
        if isinstance(key, str):
            key = Function(key, [])
        if isinstance(value, str):
            value = String(value)
        if isinstance(value, int):
            value = Number(value)
        self._factbase.add(AttributeDao(Raw(cid), Raw(key), Raw(value)))

    def add_attribute_direct(self, new_attribute):
        """
        Directly adds an attribute.
        """
        self._factbase.add(new_attribute)

    def get_elements(self):
        """
        Get all elements.
        """
        return self._factbase.query(ElementDao).all()

    def get_attributes(self, key=None):
        """
        Get all attributes.
        """
        q = self._factbase.query(AttributeDao)
        if key is not None:
            q.where(AttributeDao.key == key)
        return q.all()

    def get_callbacks(self):
        """
        Get all callbacks.
        """
        return self._factbase.query(WhenDao).all()

    def get_attributes_grouped(self):
        """
        Get all attributes grouped by element id.
        """
        return self._factbase.query(AttributeDao).group_by(AttributeDao.id).all()

    def get_callbacks_grouped(self):
        """
        Get all callbacks grouped by element id.
        """
        return self._factbase.query(WhenDao).group_by(WhenDao.id).all()

    def get_attributes_for_element_id(self, element_id):
        """
        Get all attributes for one element id.
        """
        return (
            self._factbase.query(AttributeDao)
            .where(AttributeDao.id == element_id)
            .all()
        )

    def get_callbacks_for_element_id(self, element_id):
        """
        Get all callbacks for one element id.
        """
        return self._factbase.query(WhenDao).where(WhenDao.id == element_id).all()

    def replace_attribute(self, old_attribute, new_attribute):
        """
        Replaces the old_attribute with the new_attribute.
        """
        self._factbase.remove(old_attribute)
        self._factbase.add(new_attribute)

    @classmethod
    def symbols_to_facts(cls, symbols):
        """
        Converts a iterable symbols to a string of facts.
        """
        return "\n".join([str(s) + "." for s in symbols])

    def replace_images_with_b64(self, image_attribute_key="image"):
        """
        Replaces all images in the ui-state by b64
        """
        attributes = list(self.get_attributes())
        for attribute in attributes:
            if str(attribute.key) != image_attribute_key:
                continue

            attribute_value = StandardTextProcessing.parse_string_with_quotes(
                str(attribute.value)
            )

            if os.path.isfile(attribute_value):
                with open(attribute_value, "rb") as image_file:
                    encoded_string = image_to_b64(image_file.read())
                    new_attribute = AttributeDao(
                        Raw(parse_term(str(attribute.id))),
                        Raw(Function(str(attribute.key), [])),
                        Raw(String(str(encoded_string))),
                    )
                    self.replace_attribute(attribute, new_attribute)



# pylint: disable=R0801
"""
Contains the DropdownmenuItem class.
"""

from clinguin.utils.attribute_types import StringType, SymbolType

from ..tkinter_utils import AttributeNames, CallBackDefinition, CallbackNames
from .root_cmp import RootCmp


class DropdownmenuItem(RootCmp):
    """
    Is an item of a dropdown, e.g. for the dropdownmenu countries, germany would be a dropdownmenu-item.
    """

    def _init_element(self, elements):
        parent = elements[str(self._parent)]
        if hasattr(parent, "get_menu"):
            menu = parent.get_menu()
        else:
            error_string = (
                "Parent of dropdown menu item " + self._id + " is not a dropdown menu."
            )
            self._logger.error(error_string)
            raise Exception(error_string)

        return menu

    @classmethod
    def _get_attributes(cls, attributes=None):
        if attributes is None:
            attributes = {}

        attributes[AttributeNames.label] = {"value": "", "value_type": StringType}

        return attributes

    @classmethod
    def _get_callbacks(cls, callbacks=None):
        if callbacks is None:
            callbacks = {}

        callbacks[CallbackNames.click] = {
            "operation": None,
            "operation_type": SymbolType,
        }

        return callbacks

    def _define_click_event(self, elements, key=CallbackNames.click):
        if self._callbacks[key]:
            self._element["menu"].add_command(
                label=self._attributes[AttributeNames.label]["value"],
                command=CallBackDefinition(
                    self._id,
                    self._parent,
                    self._callbacks[key]["operation"],
                    elements,
                    self._dropdownmenuitem_click,
                ),
            )

    def _dropdownmenuitem_click(self, cid, parent_id, click_operation, elements):
        parent = elements[str(parent_id)]
        if hasattr(parent, "get_variable"):
            variable = getattr(parent, "get_variable")()
            variable.set(cid)
            if click_operation is not None:
                self._base_engine.post_with_operation(click_operation)
        else:
            self._logger.warning(
                "Could not set variable for dropdownmenu. Item id: %s, dropdown-menu-id: %s",
                str(cid),
                str(parent_id),
            )

    def forget_children(self, elements):
        pass  # pylint: disable=W0107



# pylint: disable=E1101,R0801
"""
This module contains the ConfigureTextElementSize class.
"""
from tkinter import font

from clinguin.utils.attribute_types import IntegerType

from .attribute_names import AttributeNames
from .extension_class import ExtensionClass


class ConfigureTextElementSize(ExtensionClass):
    """
    If a element is a subtype of the configure-element-size class the size of the element can be adjusted.
    It is required that any subtype of this type has the ''_configure_text_element_size'' variable
    (and must be set appropriately). Further any subtype MUST also be a subtype of ''configure_font''
    """

    def __init__(self):
        self._configure_text_element_size = None

    @classmethod
    def get_attributes(cls, attributes=None):
        if attributes is None:
            attributes = {}

        attributes[AttributeNames.height] = {"value": 0, "value_type": IntegerType}
        attributes[AttributeNames.width] = {"value": 0, "value_type": IntegerType}

        return attributes

    def _set_size(self, elements):  # pylint: disable=W0613
        height = self._attributes[AttributeNames.height]["value"]
        width = self._attributes[AttributeNames.width]["value"]

        self._element.pack_propagate(0)

        used_font = self._configure_text_element_size["font"]  # pylint: disable=E1136
        afont = font.Font(
            family=used_font, size=self._attributes[AttributeNames.font_size]["value"]
        )
        text = self._attributes[AttributeNames.label]["value"]

        splits = text.split("\n")

        if width == 0:
            longest_line = max(splits, key=len)
            width = afont.measure(longest_line)  # Get width of text in pixels

        if height == 0:
            number_lines = len(splits)
            single_line_height = afont.metrics("linespace")  # Get height of one line
            height = number_lines * single_line_height

        if height > 0:
            self._element.configure(height=int(height))

        if width > 0:
            self._element.configure(width=int(width))

        if height < 0:
            self._logger.warning(
                "Height of " + self._id + " has illegal value (" + str(height) + ")"
            )

        if width < 0:
            self._logger.warning(
                "Width of " + self._id + " has illegal value (" + str(height) + ")"
            )



"""
This module contains the ImageType class.
"""

import base64

from .type import Type
from .utils.standard_text_processing import StandardTextProcessing


class ImageType(Type):
    """
    The ImageType shall be used when the image is transferred from the backend to the frontend
    via a Base64 encoded string.
    """

    @classmethod
    def parse(cls, parse_input: str, logger):
        parsed_string = StandardTextProcessing.parse_string_with_quotes(parse_input)

        try:
            image_initial_bytes = parsed_string.encode("utf-8")
            base64.b64decode(image_initial_bytes)
        except Exception as ex:
            logger.error("Sent image is not base64 encoded.")
            raise Exception("Sent image is not base64 encoded.") from ex

        return parsed_string

    @classmethod
    def description(cls):
        return "Image, needs to be a base64 encoded image, given as a string."



from utils_test_utils import UtilsTestUtils


class TestSudoku:
    def setup_method(self, test_method):
        domain_files = [
            "examples/clingo/sudoku/instance.lp",
            "examples/clingo/sudoku/encoding.lp",
        ]
        ui_files = ["examples/clingo/sudoku/ui.lp"]

        self.p, self.uvicorn_url = UtilsTestUtils.start_server(domain_files, ui_files)

    def teardown_method(self, test_method):
        UtilsTestUtils.shutdown_server(self.p)

    # TODO -> Update tests according to ''DO' syntax (or use sudoku_2)
    """
    def test_startup_get(self):
        uri = f"{self.uvicorn_url}"
        received_by_postman = str(Sudoku.get_reference_json())

        UtilsTestUtils.assert_get_request(uri, received_by_postman)


    def test_startup_next(self):
        uri = f"{self.uvicorn_url}"
        received_by_postman = str(Sudoku.get_reference_json())

        UtilsTestUtils.assert_get_request(uri, received_by_postman)

        data = '{"function":"next_solution"}'
        uri = f"{self.uvicorn_url}/backend"
        received_by_postman = str(Sudoku.post_p_1())

        UtilsTestUtils.assert_post_request(uri, received_by_postman, data)

        data = '{"function":"next_solution"}'
        uri = f"{self.uvicorn_url}/backend"
        received_by_postman = str(Sudoku.post_p_2())

        UtilsTestUtils.assert_post_request(uri, received_by_postman, data)

        data = '{"function":"clear_assumptions"}'
        uri = f"{self.uvicorn_url}/backend"
        received_by_postman = str(Sudoku.post_p_3())

        UtilsTestUtils.assert_post_request(uri, received_by_postman, data)

        uri = f"{self.uvicorn_url}"
        received_by_postman = str(Sudoku.get_reference_json())

        UtilsTestUtils.assert_get_request(uri, received_by_postman)

    def test_assumptions(self):
        uri = f"{self.uvicorn_url}"
        received_by_postman = str(Sudoku.get_reference_json())

        UtilsTestUtils.assert_get_request(uri, received_by_postman)

        post_calls = [
            ("add_assumption(sudoku(2,2,5),true)", Sudoku.post_assumption_1),
            ("add_assumption(sudoku(5,8,9),true)", Sudoku.post_assumption_2),
            ("add_assumption(sudoku(8,8,8),true)", Sudoku.post_assumption_3),
            ("add_assumption(sudoku(2,8,7),true)", Sudoku.post_assumption_4),
            ("add_assumption(sudoku(6,8,4),true)", Sudoku.post_assumption_5),
            ("add_assumption(sudoku(3,8,3),true)", Sudoku.post_assumption_6),
            ("next_solution", Sudoku.post_assumption_7),
            ("next_solution", Sudoku.post_assumption_8),
            ("clear_assumptions", Sudoku.post_assumption_9),
        ]

        for post_call in post_calls:
            data = '{"function":"' + post_call[0] + '"}'
            uri = f"{self.uvicorn_url}/backend"
            received_by_postman = str(post_call[1]())

            UtilsTestUtils.assert_post_request(uri, received_by_postman, data)

        uri = f"{self.uvicorn_url}"
        received_by_postman = str(Sudoku.get_reference_json())

        UtilsTestUtils.assert_get_request(uri, received_by_postman)
    """



"""
The utils module of clinguin - contains some useful tools for clinguin.
"""

from .attribute_types.utils.standard_text_processing import StandardTextProcessing
from .case_converter import CaseConverter, image_to_b64
from .custom_args import CustomArgs
from .errors import NoModelError, get_server_error_alert
from .logger import Logger

__all__ = [
    Logger.__name__,
    CustomArgs.__name__,
    CaseConverter.__name__,
    NoModelError.__name__,
    StandardTextProcessing.__name__,
    get_server_error_alert.__name__,
    image_to_b64.__name__,
]



"""
A clingo transformer that checks if files are using a given predicate signature
"""

import clingo.ast as _ast


class UsesSignatureTransformer(_ast.Transformer):
    """
    A transformer that checks if files are using a given predicate signature
    """

    def __init__(self, name: str, arity: int) -> None:
        self.name = name
        self.arity = arity
        self.contained = False

    def visit_Function(self, node: _ast.AST) -> _ast.AST:
        """
        Visits a function node and checks if it is the predicate signature we are looking for
        """
        if node.name == self.name and len(node.arguments) == self.arity:
            self.contained = True
        return node

    def parse_files(self, paths) -> str:
        """
        Parses the files and returns a string with the transformed program.
        """
        # pylint: disable=unnecessary-lambda
        _ast.parse_files([str(p) for p in paths], lambda stm: self(stm))



"""
This module contains the StandardTextProcessing class.
"""


class StandardTextProcessing:
    """
    The sole method parse_string_with_quotes removes unecessary quotes at the beginning and the end of a string.
    """

    @classmethod
    def parse_string_with_quotes(cls, text):
        """
        Parses a string with quotes inside and returns the string without quotes.
        Additionally it correctly parses newlines to ''correct'' newlines.
        """
        if len(text) > 0:
            if text[0] == '"':
                text = text[1:]
        if len(text) > 0:
            if text[len(text) - 1] == '"':
                text = text[:-1]

        text = text.replace("\\n", "\n")

        return text



from utils_test_utils import UtilsTestUtils


class TestClingraph:
    def setup_method(self, test_method):
        domain_files = ["examples/clingraph/coloring/encoding.lp"]
        ui_files = ["examples/clingraph/coloring/ui.lp"]

        clingraph = [
            "--backend",
            "ClingraphBackend",
            "--clingraph-files",
            "examples/clingraph/coloring/viz.lp",
            "--engine",
            "neato",
        ]

        self.p, self.uvicorn_url = UtilsTestUtils.start_server(
            domain_files, ui_files, optional=clingraph
        )

    def teardown_method(self, test_method):
        UtilsTestUtils.shutdown_server(self.p)

    # TODO -> Fix non-determinism of clingraph/graphviz!
    # Not working due to non-deterministic nature of graphviz
    # Maybe further investigation in the future ...
    """
    def test_startup_get(self):

        uri = f"{self.uvicorn_url}"
        received_by_postman = str(Coloring.get_reference_json())

        UtilsTestUtils.assert_get_request(uri, received_by_postman)

    def test_assumptions(self):

        uri = f"{self.uvicorn_url}"
        received_by_postman = str(Coloring.get_reference_json())

        UtilsTestUtils.assert_get_request(uri, received_by_postman)

        post_calls = [
            ("add_assumption(assign(1,blue),true)", Coloring.post_assumption_1),
            ("add_assumption(assign(2,green),true)", Coloring.post_assumption_2),
            ("add_assumption(assign(5,blue),true)", Coloring.post_assumption_3)
        ]

        for post_call in post_calls:

            data = '{"function":"' + post_call[0] + '"}'
            uri = f"{self.uvicorn_url}/backend"
            received_by_postman = str(post_call[1]())

            UtilsTestUtils.assert_post_request(uri, received_by_postman, data)
    """



"""
Responsible for parsing the command line attributes
"""

import argparse
import importlib
import inspect
import os
import sys
import textwrap
import traceback

from .client import AbstractFrontend
from .server.application.backends.clingo_backend import ClingoBackend
from .show_frontend_syntax_enum import ShowFrontendSyntaxEnum

if sys.version_info[1] < 8:
    import importlib_metadata as metadata  # nocoverage
else:
    from importlib import metadata  # nocoverage

VERSION = metadata.version("clinguin")


class ArgumentParser:
    """
    ArgumentParser-Class, Responsible for parsing the command line attributes
    """

    default_backend_exec_string = "from .server.application.backends import *"
    default_frontend_exec_string = "from .client.presentation.frontends import *"

    default_backend = "ClingoBackend"
    default_frontend = "AngularFrontend"

    def __init__(self) -> None:
        self.frontend_name = None
        self.frontend = None

        self.titles = {
            "client": self._client_title,
            "server": self._server_title,
            "client-server": self._client_server_title,
        }
        self.descriptions = {
            "client": "Start a client process that will render a UI.",
            "server": "Start server process making endpoints available for a client.",
            "client-server": "Start client and a server processes.",
        }
        self.backend_name = None
        self.backend = None
        self._show_frontend_syntax = ShowFrontendSyntaxEnum.NONE

    def parse(self, process, string_args):
        """
        After initialization of the ArgumentParser call this function to parse the arguments.
        """
        self._parse_custom_classes(string_args)

        parser = argparse.ArgumentParser(
            description=self._clinguin_description(process),
            add_help=True,
            formatter_class=argparse.RawTextHelpFormatter,
        )
        parser.add_argument(
            "--version", "-v", action="version", version=f"%(prog)s {VERSION}"
        )
        subparsers = parser.add_subparsers(
            title="Process type",
            description="The type of process to start: a client (UI) a server (Backend) or both",
            dest="process",
        )
        self._create_client_subparser(subparsers)
        self._create_server_subparser(subparsers)
        self._create_client_server_subparser(subparsers)

        args = parser.parse_args(string_args)

        self._add_selected_backend(args)

        return args

    def _add_selected_backend(self, args):
        args.backend = self.backend
        args.frontend = self.frontend

    @property
    def _clinguin_title(self):
        return """
              ___| (_)_ __   __ _ _   _(_)_ __
             / __| | | '_ \\ / _` | | | | | '_ \\
            | (__| | | | | | (_| | |_| | | | | |
             \\___|_|_|_| |_|\\__, |\\__,_|_|_| |_|
                            |___/
            """

    @property
    def _client_server_title(self):
        return """
             _ | o  _  ._  _|_     _  _  ._     _  ._
            (_ | | (/_ | |  |_    _> (/_ |  \\/ (/_ |

            """

    @property
    def _client_title(self):
        return """
                       _ | o  _  ._  _|_
                      (_ | | (/_ | |  |_

            """

    @property
    def _server_title(self):
        return """
                       _  _  ._     _  ._
                      _> (/_ |  \\/ (/_ |

            """

    def _clinguin_description(self, process):
        description = (
            "Clinguin is a GUI language extension for a logic program that uses Clingo."
        )
        if process not in ["server", "client", "client-server"]:
            ascci = f"{self._clinguin_title}"
            return_value = f"{inspect.cleandoc(ascci)}\n\n{description}"
        else:
            ascci = f"{self._clinguin_title}{self.titles[process]}"
            return_value = f"{inspect.cleandoc(ascci)}\n\n{description}\n{self.descriptions[process]}"

        return return_value

    def _import_classes(self, path):
        if os.path.isfile(path):
            sys.path.append(os.path.dirname(path))
            self._import_files_from_path_array([path])
        else:
            sys.path.append(path)
            self._recursive_import(path, "", "")

    def _import_files_from_path_array(self, file_paths, module=""):
        for file_path in file_paths:
            base = os.path.basename(file_path)
            file_name = os.path.splitext(base)[0]
            ending = os.path.splitext(base)[1]
            if ending == ".py":
                if module != "":
                    try:
                        importlib.import_module(module + "." + file_name)
                    except Exception as ex:
                        raise Exception(
                            "Could not import module: " + module + "." + file_name
                        ) from ex
                else:
                    importlib.import_module(file_name)

    def _recursive_import(self, full_path, rec_path, module):
        folder_paths = []
        file_paths = []

        try:
            for entity in os.scandir(os.path.join(full_path, rec_path)):
                if entity.is_dir():
                    folder_paths.append(entity.path)
                elif entity.is_file():
                    file_paths.append(entity.path)
        except Exception as ex:
            print("<<<BEGIN-STACK-TRACE>>>")
            traceback.print_exc()
            print("<<<END-STACK-TRACE>>>")
            raise Exception(
                "Could not find path for importing libraries: "
                + os.path.join(full_path, rec_path)
                + ". Therefore program is terminating now (full stacktrace is printed below)."
            ) from ex

        self._import_files_from_path_array(file_paths)

        for folder_path in folder_paths:
            base = os.path.basename(folder_path)
            new_module = module
            if new_module != "":
                new_module = new_module + "." + base
            else:
                new_module = base

            self._recursive_import(full_path, os.path.join(rec_path, base), new_module)

    def _parse_custom_classes(self, str_args):
        custom_imports_parser = argparse.ArgumentParser(add_help=False)
        self._add_default_arguments_to_backend_parser(custom_imports_parser)
        self._add_default_arguments_to_client_parser(custom_imports_parser)

        args, _ = custom_imports_parser.parse_known_args(str_args)

        self.frontend_name = args.frontend
        self.backend_name = args.backend
        if args.custom_classes:
            self._import_classes(args.custom_classes)

        exec(ArgumentParser.default_backend_exec_string)
        exec(ArgumentParser.default_frontend_exec_string)

        if args.frontend_syntax and not args.frontend_syntax_full:
            self._show_frontend_syntax = ShowFrontendSyntaxEnum.SHOW
        elif args.frontend_syntax_full:
            self._show_frontend_syntax = ShowFrontendSyntaxEnum.FULL

    def _create_client_subparser(self, subparsers):
        parser_client = subparsers.add_parser(
            "client",
            help=self.descriptions["client"],
            description=self._clinguin_description("client"),
            add_help=True,
            formatter_class=argparse.RawTextHelpFormatter,
        )

        self._add_log_arguments(
            parser_client, abbrevation="C", logger_name="clinguin_client"
        )

        parser_client.add_argument(
            "--server-port",
            type=int,
            default=8000,
            help="Define the Port of the server, default value is 8000.",
        )

        parser_client.add_argument(
            "--server-url",
            type=str,
            default="http://localhost",
            help="Define the URL of the server, default value is 'http://localhost'.",
        )

        self._add_default_arguments_to_client_parser(parser_client)
        self.frontend = self._select_subclass_and_add_custom_arguments(
            parser_client,
            AbstractFrontend,
            self.frontend_name,
            ArgumentParser.default_frontend,
        )

        return parser_client

    def _create_server_subparser(self, subparsers):
        parser_server = subparsers.add_parser(
            "server",
            help=self.descriptions["server"],
            description=self._clinguin_description("server"),
            add_help=True,
            formatter_class=argparse.RawTextHelpFormatter,
        )

        self._add_log_arguments(
            parser_server, abbrevation="S", logger_name="clinguin_server"
        )
        self._add_default_arguments_to_backend_parser(parser_server)
        self.backend = self._select_subclass_and_add_custom_arguments(
            parser_server,
            ClingoBackend,
            self.backend_name,
            ArgumentParser.default_backend,
        )

        return parser_server

    def _create_client_server_subparser(self, subparsers):
        parser_server_client = subparsers.add_parser(
            "client-server",
            help=self.descriptions["client-server"],
            description=self._clinguin_description("client-server"),
            add_help=True,
            formatter_class=argparse.RawTextHelpFormatter,
        )

        self._add_log_arguments(
            parser_server_client,
            abbrevation="C",
            logger_name="clinguin_client",
            display_name="client-",
        )
        self._add_log_arguments(
            parser_server_client,
            abbrevation="S",
            logger_name="clinguin_server",
            display_name="server-",
        )

        self._add_default_arguments_to_client_parser(parser_server_client)
        self.frontend = self._select_subclass_and_add_custom_arguments(
            parser_server_client,
            AbstractFrontend,
            self.frontend_name,
            ArgumentParser.default_frontend,
        )

        self._add_default_arguments_to_backend_parser(parser_server_client)
        self.backend = self._select_subclass_and_add_custom_arguments(
            parser_server_client,
            ClingoBackend,
            self.backend_name,
            ArgumentParser.default_backend,
        )

        return parser_server_client

    def _add_default_arguments_to_backend_parser(self, parser):
        sub_classes = [ClingoBackend] + self._get_sub_classes(ClingoBackend)
        sub_class_as_options = "|".join([s.__name__ for s in sub_classes])
        sub_classes_str = "=>  Available options: {" + sub_class_as_options + "}"
        parser.add_argument(
            "--backend",
            type=str,
            help=textwrap.dedent(
                f"""\
                Optionally specify which backend to use using the class name.
                {sub_classes_str}
                """
            ),
            metavar="",
        )
        parser.add_argument(
            "--custom-classes", type=str, help="Path to custom classes.", metavar=""
        )
        parser.add_argument(
            "--server-port",
            type=int,
            default=8000,
            help="Define the server side port, default value is 8000.",
        )

    def _add_default_arguments_to_client_parser(self, parser):
        sub_classes = self._get_sub_classes(AbstractFrontend)
        sub_class_as_options = "|".join([s.__name__ for s in sub_classes])
        sub_classes_str = "=>  Available options: {" + sub_class_as_options + "}"
        parser.add_argument(
            "--frontend",
            type=str,
            help=textwrap.dedent(
                f"""\
                Optionally specify which frontend to use using the class name.
                {sub_classes_str}
                """
            ),
            metavar="",
        )
        parser.add_argument(
            "--frontend-syntax",
            action="store_true",
            help="Show available commands for the GUI.",
        )
        parser.add_argument(
            "--frontend-syntax-full",
            action="store_true",
            help="Show available commands for the GUI and shows available value-types.",
        )

    def _add_log_arguments(
        self, parser, abbrevation="", logger_name="", display_name=""
    ):
        group = parser.add_argument_group(display_name + "logger")
        group.add_argument(
            "--" + display_name + "log-disable-shell",
            action="store_true",
            help="Disable shell logging",
        )
        group.add_argument(
            "--" + display_name + "log-enable-file",
            action="store_true",
            help="Disable file logging",
        )
        group.add_argument(
            "--" + display_name + "logger-name",
            type=str,
            help="Set logger name",
            metavar="",
            default=logger_name,
        )
        group.add_argument(
            "--" + display_name + "log-level",
            type=str,
            help="Log level",
            metavar="",
            choices=["DEBUG", "INFO", "ERROR", "WARNING"],
            default="INFO",
        )
        group.add_argument(
            "--" + display_name + "log-format-shell",
            type=str,
            help="Log format shell",
            metavar="",
            default="[" + str(abbrevation) + "] %(levelname)s: %(message)s",
        )
        group.add_argument(
            "--" + display_name + "log-format-file",
            type=str,
            help="Log format file",
            metavar="",
            default="%(levelname)s: %(message)s",
        )

    def _get_sub_classes(self, cur_class):
        sub_classes = cur_class.__subclasses__()
        recursive = []
        for sub_class in sub_classes:
            recursive.extend(self._get_sub_classes(sub_class))

        return sub_classes + recursive

    def _select_subclass_and_add_custom_arguments(
        self, parser, parent, class_name, default_class
    ):
        sub_classes = [parent] + self._get_sub_classes(parent)

        selected_class = None

        for sub_class in sub_classes:
            full_class_name = sub_class.__name__
            selected_by_default = not class_name and full_class_name == default_class
            selected = full_class_name == class_name
            if selected_by_default or selected:
                group = parser.add_argument_group(full_class_name)
                sub_class.register_options(group)

                should_show_frontend_syntax = self._show_frontend_syntax in [
                    ShowFrontendSyntaxEnum.SHOW,
                    ShowFrontendSyntaxEnum.FULL,
                ]

                if should_show_frontend_syntax and hasattr(
                    sub_class, "available_syntax"
                ):
                    print(sub_class.available_syntax(self._show_frontend_syntax))
                    sys.exit()

                selected_class = sub_class

        if not selected_class:
            raise RuntimeError(f"Invalid class name provided: {class_name}")

        return selected_class



"""
Defines the possibilities how to show the syntax.
"""

from enum import Enum, auto


class ShowFrontendSyntaxEnum(Enum):
    """
    Defines the possibilities how to show the syntax.
    """

    NONE = auto()
    SHOW = auto()
    FULL = auto()



"""
Module contains the CaseConverter class.
"""

import base64


class CaseConverter:
    """
    The CaseConverter class provides functionality to convert strings from one type-case
    (like snake_case) to another (like camelCase).

    Methods:
        snake_case_to_camel_case(snake_case : str) -> str
    """

    @classmethod
    def snake_case_to_camel_case(cls, snake_case):
        """
        Converts a snake case name to a camel case one.
        """
        components = snake_case.split("_")

        return components[0] + "".join(x.title() for x in components[1:])


def image_to_b64(img):
    """
    Encodes an image in base 64
    """
    encoded = base64.b64encode(img)
    decoded = encoded.decode("utf-8")
    return decoded



"""
Module that contains the Explanation Backend.
"""

import textwrap
from functools import cached_property

from clingexplaid.mus import CoreComputer
from clingexplaid.preprocessors import AssumptionPreprocessor, FilterSignature
from clingo.script import enable_python

from clinguin.server.application.backends.clingo_backend import (
    ClingoBackend,
)
from clinguin.utils.annotations import extends


enable_python()


class ExplanationBackend(ClingoBackend):
    """
    Extends ClingoBackend. This backend will treat an UNSAT result by adding the
    Minimal Unsatisfiable Core (MUC) to the domain-state, thus allowing the UI to show
    the faulty assumptions.
    """

    # ---------------------------------------------
    # Properties
    # ---------------------------------------------

    @property
    def _assumption_list(self):
        """
        Gets the set of assumptions used for solving.
        It includes the assumptions from the assumption signatures provided.

        Warning:

            Overwrites :meth:`ClingoBackend._assumption_list`
        """
        return self._assumptions.union(self._assumptions_from_signature)

    # ---------------------------------------------
    # Initialization
    # ---------------------------------------------

    @extends(ClingoBackend)
    def _init_command_line(self):
        """
        Sets the assumption signature and the transformer used for the input files

        Attributes:
            _assumption_sig (List[Tuple[str, int]]): The list of assumption signatures
            _assumption_transformer (clingexplaid.AssumptionTransformer): The transformer used for the input files
        """
        super()._init_command_line()
        # pylint: disable= attribute-defined-outside-init
        self._assumption_sig = []
        for a in self._args.assumption_signature or []:
            try:
                name, arity = a.split(",")
                self._assumption_sig.append((name, int(arity)))
            except Exception as ex:
                raise ValueError(
                    "Argument assumption_signature must have format name,arity"
                ) from ex

    def _create_ctl(self) -> None:
        """
        Initializes the control object (domain-control).
        It is used when the server is started or after a restart.

        See Also:
            :func:`~_load_file`
        """
        super()._create_ctl()
        filters = {FilterSignature(s, a) for s, a in self._assumption_sig}
        # pylint: disable=attribute-defined-outside-init
        self._assumption_transformer = AssumptionPreprocessor(
            control=self._ctl,
            filters=filters,
        )

    @extends(ClingoBackend)
    def _init_ds_constructors(self):
        super()._init_ds_constructors()
        self._add_domain_state_constructor("_ds_mus")

    @extends(ClingoBackend)
    def _load_file(self, f):
        """
        Loads a file into the control. Transforms the program to add choices around assumption signatures.

        Arguments:
            f (str): The file path

        """
        self._logger.debug("Assumption transformer will process file %s", f)
        self._assumption_transformer.process_files([f])

    # ---------------------------------------------
    # Solving
    # ---------------------------------------------
    @extends(ClingoBackend)
    def _ground(self, program="base", arguments=None):
        """
        Sets the list of assumptions that were taken from the input files using the assumption_signature.

        Attributes:
            _assumptions_from_signature (Set[Tuple(str,bool)]): The set of assumptions from the assumption signatures
            arguments (list, optional): The list of arguments to ground the program. Defaults to an empty list.
        """
        super()._ground(program, arguments)
        # pylint: disable= attribute-defined-outside-init
        self._assumptions_from_signature = list(
            self._assumption_transformer.assumptions
        )

    # ---------------------------------------------
    # Class methods
    # ---------------------------------------------

    @classmethod
    @extends(ClingoBackend)
    def register_options(cls, parser):
        """
        Registers command line options for ClingraphBackend.
        """
        ClingoBackend.register_options(parser)

        parser.add_argument(
            "--assumption-signature",
            help=textwrap.dedent(
                """\
                            Signatures that will be considered as true assumptions. Must be have format name,arity"""
            ),
            nargs="+",
            metavar="",
        )

    # ---------------------------------------------
    # Domain state
    # ---------------------------------------------

    @cached_property
    def _ds_mus(self):
        """
        Adds information about the Minimal Unsatisfiable Set (MUS)
        Includes predicate ``_clinguin_mus/1`` for every assumption in the MUC
        It uses a cache that is erased after an operation makes changes in the control.
        """
        prg = "#defined _clinguin_mus/1. "
        if self._unsat_core is not None:
            self._logger.info("UNSAT Answer, will add explanation")
            mus_core = CoreComputer(self._ctl, self._assumption_list).shrink()
            prg += " ".join([f"_clinguin_mus({str(s)})." for s, _ in mus_core]) + "\n"
        return prg



"""
Module that contains the ChildLayoutType.
"""

from enum import auto

from .enum import EnumType
from .utils.standard_text_processing import StandardTextProcessing


class ChildLayoutType(EnumType):
    """
    The ChildLayoutType is an Enum Type, which can be used to specify the layout of the children.
    """

    FLEX = auto()
    GRID = auto()
    ABSSTATIC = auto()
    RELSTATIC = auto()

    @classmethod
    def parse(cls, parse_input: str, logger):
        parsed_string = (
            StandardTextProcessing.parse_string_with_quotes(parse_input)
        ).lower()

        return_value = None

        if parsed_string == cls.FLEX.name.lower():
            return_value = cls.FLEX
        elif parsed_string == cls.GRID.name.lower():
            return_value = cls.GRID
        elif parsed_string == cls.ABSSTATIC.name.lower():
            return_value = cls.ABSSTATIC
        elif parsed_string == cls.RELSTATIC.name.lower():
            return_value = cls.RELSTATIC
        else:
            logger.error("Could not parse " + parsed_string + " to child_layout type.")
            raise Exception(
                "Could not parse " + parsed_string + " to child_layout type."
            )

        return return_value

    @classmethod
    def description(cls):
        return (
            "For the child-layout four different options exists: flex (default, tries to do it automatically),"
            + "grid (grid-like-specification), absstatic (if one wants to specify the position with absolute-pixel"
            + "coordinates) and relstatic (if one wants to specify the position with relative-pixel coordinates"
            + "(from 0 to 100 percent, where 0 means left/top and 100 means right/bottom)). They can either be"
            + "specified via a clingo symbol or via a string (string is case-insensitive)."
        )



# pylint: disable=R1711
"""
Contains the RootCmp class.
"""
import logging

from clinguin.utils.attribute_types import ChildLayoutType, StringType, SymbolType

from ..tkinter_utils import ExtensionClass


class RootCmp:
    """
    Every tkinter element must be a subtype of the RootCmp.
    It further features standard implementations of various methods,
    therefore one must just implement a handful of methods if one implements a new element,
    these are (see e.g. the button.py for a sample implementation):
    - _get_attributes(attributes = None)
    - _get_callbacks(callbacks = None)
    - Every method, which name starts with ''_set'' will be executed and are designed to be used
    for setting the attributes.
    - Every method, which name starts with ''_define'' will be executed and are designet to specify callbacks.
    """

    def __init__(self, args, cid, parent, attributes, callbacks, base_engine):
        self._logger = logging.getLogger(args.log_args["name"])
        self._id = cid
        self._parent = parent
        self._json_attributes = attributes
        self._json_callbacks = callbacks
        self._base_engine = base_engine
        self._element = None

        self._attributes = None
        self._callbacks = None

    def get_id(self):
        """
        Get the id of the element.
        """
        return self._id

    def get_parent(self):
        """
        Get the parent of the element.
        """
        return self._parent

    def get_element(self):
        """
        Get the tkinter object of the element.
        """
        return self._element

    def get_attributes_list(self):
        """
        Get the attributes list of instance.
        """
        return self._attributes

    @classmethod
    def get_attributes(cls):
        """
        Get attributes list of class (which attributes has this class in general).
        """
        attributes = {}
        for base in cls.__bases__:
            if issubclass(base, ExtensionClass):
                base.get_attributes(attributes)

        return cls._get_attributes(attributes)

    @classmethod
    def _get_attributes(cls, attributes=None):  # pylint: disable=W0613
        return {}

    @classmethod
    def get_callbacks(cls):
        """
        Get callbacks list of class (which callbacks has this class in general).
        """
        callbacks = {}
        for base in cls.__bases__:
            if issubclass(base, ExtensionClass):
                base.get_callbacks(callbacks)

        return cls._get_callbacks(callbacks)

    @classmethod
    def _get_callbacks(cls, callbacks=None):  # pylint: disable=W0613
        return {}

    def add_component(self, elements):
        """
        Add the current (self) component to elements.
        """
        self._element = self._init_element(elements)  # pylint: disable=E1128
        self._attributes = self.__class__.get_attributes()
        self._callbacks = self.__class__.get_callbacks()

        self._fill_attributes()
        self._fill_callbacks()

        self._exec_attributes(elements)
        self._exec_actions(elements)

        self._add_component_to_elements(elements)

    def _init_element(self, elements):
        self._logger.debug(str(elements))
        return None

    def _fill_attributes(self):
        for attribute in self._json_attributes:
            key = attribute["key"]
            value = attribute["value"]
            if key in self._attributes and "value_type" in self._attributes[key]:
                value_type = self._attributes[key]["value_type"]
            else:
                value_type = StringType
            if key in self._attributes and "value" in self._attributes[key]:
                self._attributes[key]["value"] = value_type.parse(value, self._logger)
            else:
                self._logger.warning(
                    "Undefined Command: " + key + " for element: " + attribute["id"]
                )

    def _fill_callbacks(self):
        for callback in self._json_callbacks:
            if callback["action"] not in ["call", "callback"]:
                self._logger.warning(
                    "Only interaction type call and callback are available in the Tkinter frontend.\
                    Interactivity: '%s' was ignored",
                    callback["action"],
                )
            key = callback["event"]
            value = callback["operation"]
            if key in self._callbacks and "operation_type" in self._callbacks[key]:
                value_type = self._callbacks[key]["operation_type"]
            else:
                value_type = SymbolType

            if key in self._callbacks and "operation" in self._callbacks[key]:
                self._callbacks[key]["operation"] = value_type.parse(
                    value, self._logger
                )
            else:
                self._logger.warning(
                    "Undefined Command: %s, or operation item missing in command.", key
                )

    def _get_methods(self, start_string):
        object_methods = []
        for method_name in dir(self):
            if method_name.startswith(start_string) and callable(
                getattr(self, method_name)
            ):
                object_methods.append(getattr(self, method_name))

        return object_methods

    def _exec_attributes(self, elements):
        attribute_methods = self._get_methods("_set")
        for set_attribute_method in attribute_methods:
            set_attribute_method(elements)

    def _exec_actions(self, elements):
        callback_methods = self._get_methods("_define")
        for define_callback_method in callback_methods:
            define_callback_method(elements)

    def _add_component_to_elements(self, elements):
        elements[str(self._id)] = self

    def forget_children(self, elements):
        """
        The current element drops all relations to its children (thereby 'forgetting' them).
        """
        if str(self._parent) in elements:
            if hasattr(elements[self._parent], "get_child_org"):
                parent_org = getattr(elements[self._parent], "get_child_org")()
                if parent_org in (
                    ChildLayoutType.FLEX,
                    ChildLayoutType.RELSTATIC,
                    ChildLayoutType.ABSSTATIC,
                ):
                    self._element.pack_forget()
                elif parent_org == ChildLayoutType.GRID:
                    self._element.grid_forget()
                else:
                    self._element.forget()
            else:
                self._element.forget()
        else:
            pass



# pylint: disable=R0903
"""
Module that contains the CallbackDao.
"""
from clorm import Predicate, RawField


class WhenDao(Predicate):
    """
    Actions define the interactivity of the UI.  Multiple actions are allowed, as explained below.

    - id: Identifier of the element that the user interacted with.
    - event: The event that is being triggered, such as click, hover,  input, etc.
            Each element type allows different events.
    - action: The action performed.
        - call: Calls the server to perform an operation.
        - update: Updates the attribute of another element without any calls to the server.
        - context: Updates the internal context that will be passed to the server on the following call actions.
                    See Context for more details.
    - operation: The operation accounts for the information that the action requires for its execution.

    """

    # pylint: disable=abstract-method

    id = RawField
    event = RawField
    action = RawField
    operation = RawField

    class Meta:
        """
        Meta class
        """

        name = "when"



# pylint: disable=W0107
"""
The module contains the CustomArgs class.
"""


class CustomArgs:
    """
    The root class of all backend- (like ClingoBackend) and gui-classes (like TkinterFrontend).
    Needed for argparse (with the method register_options one can specify command line arguments,
    that are specific to a certain class).

    Method:
        register_options (parser: argparse-object)
    """

    @classmethod
    def register_options(cls, parser):
        """
        Method that needs to be implemented by all children.
        """



"""
Module that contains the ClientBase class
"""

import logging
import time

from clinguin.client.api.api import Api
from clinguin.client.api.frontend_operation_dto import FrontendOperationDto
from clinguin.utils import CaseConverter, Logger


class ClientBase:
    """
    ClientBase is the ''base'' of the client.
    It contains the logic which is responsible for connecting to the server,
    what to do in case of errors, forward the Json to the correct GUI, etc.
    """

    endpoint_health = "health"

    def __init__(self, args):
        self._logger = logging.getLogger(Logger.client_logger_name)

        self.api = Api()
        self.connected = False

        self.solve_dto = FrontendOperationDto("solve")

        self.frontend_generator = args.frontend(self, args)

    def start_up(self):
        """
        Code executed on first startup.
        """
        self.connect()
        (status_code, response) = self.api.get("")
        if status_code == 200:
            self.draw(response["ui"])
            self.frontend_generator.draw(response["ui"]["children"][0]["id"])
        else:
            self._logger.error("Connection error, status code: %s", str(status_code))

            self.connected = False
            self.connect()

    def connect(self):
        """
        Connect/reconnect to server.
        """
        while not self.connected:
            (status_code, _) = self.api.get(self.endpoint_health)

            if status_code == 200:
                self.connected = True
            else:
                self._logger.info("Waiting for connection")
                time.sleep(1)

    def draw(self, response):
        """
        Draws the GUI.
        """
        self.base_engine(response)
        if len(response["children"]) == 0:
            raise Exception("Empty UI! check the logs and the provided files.")
        self.frontend_generator.draw_postprocessing(response["ui"]["children"][0]["id"])

    def base_engine(self, response):
        """
        Handles the response of the server to draw the GUI.
        For this it iterates through the received Json/Dict
        in a top-down preorder fashion.

        Parameters:
            response (dict): Json from which one can draw the GUI.
        """
        children = response["children"]

        for child in children:
            snake_case_name = child["type"]
            camel_case_name = CaseConverter.snake_case_to_camel_case(snake_case_name)

            method = None

            if hasattr(self.frontend_generator, camel_case_name):
                method = getattr(self.frontend_generator, camel_case_name)

            elif hasattr(self.frontend_generator, snake_case_name):
                method = getattr(self.frontend_generator, snake_case_name)

            if method and callable(method):
                method(
                    child["id"],
                    child["parent"],
                    child["attributes"],
                    child["when"],
                )
                self.base_engine(child)
            else:
                self._logger.error("Could not find element type: %s", child["type"])

    def post_with_operation(self, click_operation):
        """
        Prepare post request for API.
        """
        (status_code, json) = self.api.post(
            "backend", FrontendOperationDto(click_operation)
        )
        if status_code == 200:
            self.draw(json["ui"])
        else:
            self._logger.error("Connection error, status code: %s", str(status_code))

            self.connected = False
            self.connect()



# pylint: disable=R0801
"""
This module contains the Dropdownmenu class.
"""
import tkinter as tk

from clinguin.utils.attribute_types import (
    BooleanType,
    ColorType,
    StringType,
    SymbolType,
)

from ..tkinter_utils import AttributeNames, CallbackNames, ConfigureSize, LayoutFollower
from .root_cmp import RootCmp


class Dropdownmenu(RootCmp, LayoutFollower, ConfigureSize):
    """
    The dropdownmenu is the master component for a dropdownmenu,
    i.e. dropdownmenu-items must be children of it.
    For available attributes see syntax definition.
    Implementation wise it is similarly implemented as the Label and Button -
    to make it work for layouting, the actual dropdownmenu is hidden and the element
    is actually a tkinter frame (therefore self._element is a frame, whereas self._menu is the dropdownmenu).
    """

    def __init__(self, args, cid, parent, attributes, callbacks, base_engine):
        super().__init__(args, cid, parent, attributes, callbacks, base_engine)

        self._menu = None
        self._variable = None

    def _init_element(self, elements):
        option_menu_frame = tk.Frame(elements[str(self._parent)].get_element())

        self._variable = tk.StringVar()
        items = []
        menu = tk.OptionMenu(option_menu_frame, self._variable, "", *items)

        self._menu = menu

        return option_menu_frame

    def get_variable(self):
        """
        Gets the variable associated with the value.
        """
        return self._variable

    @classmethod
    def _get_attributes(cls, attributes=None):
        if attributes is None:
            attributes = {}

        attributes[AttributeNames.backgroundcolor] = {
            "value": "white",
            "value_type": ColorType,
        }
        attributes[AttributeNames.foregroundcolor] = {
            "value": "black",
            "value_type": ColorType,
        }
        # Interactive-Attributes
        attributes[AttributeNames.onhover] = {"value": False, "value_type": BooleanType}
        attributes[AttributeNames.onhover_background_color] = {
            "value": "white",
            "value_type": ColorType,
        }
        attributes[AttributeNames.onhover_foreground_color] = {
            "value": "black",
            "value_type": ColorType,
        }

        attributes[AttributeNames.selected] = {"value": "", "value_type": StringType}

        return attributes

    @classmethod
    def _get_callbacks(cls, callbacks=None):
        if callbacks is None:
            callbacks = {}

        callbacks[CallbackNames.clear] = {
            "operation": None,
            "operation_type": SymbolType,
        }

        return callbacks

    def _clear_select(self, cid, parent_id, click_operation, elements):
        parent = elements[str(parent_id)]
        if hasattr(parent, "get_variable"):
            variable = getattr(parent, "get_variable")()
            variable.set(cid)
            if click_operation is not None:
                self._base_engine.post_with_operation(click_operation)
        else:
            self._logger.warning(
                "Could not set variable for dropdownmenu. Item id: %s, dropdown-menu-id: %s",
                str(id),
                str(parent_id),
            )

    def _dropdown_clear(self, click_operation):
        variable = self.get_variable()
        variable.set("")
        self._base_engine.post_with_operation(click_operation)

    def _define_clear_event(
        self, elements, key=CallbackNames.clear
    ):  # pylint: disable=W0613
        if self._callbacks[key]["operation"] is None:
            return

        def change(*args):  # pylint: disable=W0613
            if self._variable.get() == "":
                self._logger.info("Will remove previous selections")
                self._dropdown_clear(self._callbacks[key]["operation"])

        self._variable.trace("w", change)

    def _set_background_color(
        self, elements, key=AttributeNames.backgroundcolor
    ):  # pylint: disable=W0613
        value = self._attributes[key]["value"]

        self._menu.config(bg=value, activebackground=value)
        self._menu["menu"].config(bg=value, activebackground=value)

    def _set_foreground_color(
        self, elements, key=AttributeNames.foregroundcolor
    ):  # pylint: disable=W0613
        value = self._attributes[key]["value"]

        self._menu.config(fg=value, activeforeground=value)
        self._menu["menu"].config(fg=value, activeforeground=value)

    def _set_on_hover(self, elements):  # pylint: disable=W0613
        on_hover = self._attributes[AttributeNames.onhover]["value"]

        on_hover_background_color = self._attributes[
            AttributeNames.onhover_background_color
        ]["value"]

        on_hover_foreground_color = self._attributes[
            AttributeNames.onhover_foreground_color
        ]["value"]

        if on_hover == "true":
            self._menu.config(
                activebackground=on_hover_background_color,
                activeforeground=on_hover_foreground_color,
            )
            self._menu["menu"].config(
                activebackground=on_hover_background_color,
                activeforeground=on_hover_foreground_color,
            )

    def _set_selected(self, elements):  # pylint: disable=W0613
        self._variable.set(self._attributes[AttributeNames.selected]["value"])

    def get_menu(self):
        """
        Returns the actual tkinter menu of this element.
        """
        return self._menu

    def _add_component_to_elements(self, elements):
        self._menu.pack(expand=True, fill="both")

        elements[str(self._id)] = self



from reference_json_output import (
    BasicTest00,
    BasicTest01,
    BasicTest02,
    BasicTest03,
    BasicTest04,
    BasicTest05,
)
from utils_test_utils import UtilsTestUtils


class TestBackendUnit_00_05:
    def setup_method(self, test_method):
        self.backend = UtilsTestUtils.instantiate_backend(test_method)

    def teardown_method(self, test_method):
        pass

    def test_basic_05(self):
        should_output = BasicTest05.get_reference_json()
        received_by_backend = self.backend.get()
        UtilsTestUtils.assert_result(should_output, received_by_backend["ui"])

    def test_basic_04(self):
        should_output = BasicTest04.get_reference_json()
        received_by_backend = self.backend.get()
        UtilsTestUtils.assert_result(should_output, received_by_backend["ui"])

    def test_basic_03_order_04(self):
        should_output = BasicTest03.get_reference_json()
        received_by_backend = self.backend.get()
        UtilsTestUtils.assert_result(should_output, received_by_backend["ui"])

        self.backend.add_assumption("p(1)", "true")
        received_by_backend = self.backend.get()
        should_output = BasicTest03.post_p_1_reference_json()
        UtilsTestUtils.assert_result(should_output, received_by_backend["ui"])

        self.backend.clear_assumptions()
        received_by_backend = self.backend.get()
        should_output = BasicTest03.post_p_3_reference_json()
        UtilsTestUtils.assert_result(should_output, received_by_backend["ui"])

    def test_basic_03_order_03(self):
        should_output = BasicTest03.get_reference_json()
        received_by_backend = self.backend.get()
        UtilsTestUtils.assert_result(should_output, received_by_backend["ui"])

        self.backend.clear_assumptions()
        received_by_backend = self.backend.get()
        should_output = BasicTest03.post_p_2_reference_json()
        UtilsTestUtils.assert_result(should_output, received_by_backend["ui"])

    def test_basic_03_order_02(self):
        should_output = BasicTest03.get_reference_json()
        received_by_backend = self.backend.get()
        UtilsTestUtils.assert_result(should_output, received_by_backend["ui"])

        self.backend.add_assumption("p(1)", "true")
        received_by_backend = self.backend.get()
        should_output = BasicTest03.post_p_1_reference_json()
        UtilsTestUtils.assert_result(should_output, received_by_backend["ui"])

    def test_basic_03_order_01(self):
        should_output = BasicTest03.get_reference_json()
        received_by_backend = self.backend.get()
        UtilsTestUtils.assert_result(should_output, received_by_backend["ui"])

    def test_basic_02(self):
        should_output = BasicTest02.get_reference_json()
        received_by_backend = self.backend.get()
        UtilsTestUtils.assert_result(should_output, received_by_backend["ui"])

    def test_basic_01_order_02(self):
        should_output = BasicTest01.get_reference_json()
        received_by_backend = self.backend.get()
        UtilsTestUtils.assert_result(should_output, received_by_backend["ui"])

        self.backend.add_assumption("p(2)", "true")
        received_by_backend = self.backend.get()
        should_output = BasicTest01.post_p_2_reference_json()
        UtilsTestUtils.assert_result(should_output, received_by_backend["ui"])

        should_output = BasicTest01.get_p_2_reference_json()
        received_by_backend = self.backend.get()
        UtilsTestUtils.assert_result(should_output, received_by_backend["ui"])

    def test_basic_01_order_01(self):
        should_output = BasicTest01.get_reference_json()
        received_by_backend = self.backend.get()
        UtilsTestUtils.assert_result(should_output, received_by_backend["ui"])

        self.backend.add_assumption("p(1)", "true")
        received_by_backend = self.backend.get()
        should_output = BasicTest01.post_p_1_reference_json()
        UtilsTestUtils.assert_result(should_output, received_by_backend["ui"])

        should_output = BasicTest01.get_p_1_reference_json()
        received_by_backend = self.backend.get()
        UtilsTestUtils.assert_result(should_output, received_by_backend["ui"])

    def test_get_method(self):
        should_output = BasicTest00.get_reference_json()
        received_by_backend = self.backend.get()
        UtilsTestUtils.assert_result(should_output, received_by_backend["ui"])



"""
Module that contains the API-Class
"""

import json
import logging
import traceback
from urllib.request import Request, URLError, urlopen

from clinguin.utils import Logger

from .frontend_operation_dto import FrontendOperationDto


class Api:
    """
    The API class is responsible for handling API calls, i.e. sending requests and receiving the responses.
    For this two methods exists - GET and POST.
    This Api sends by default requests to the url:http://127.0.0.1:8000/
    """

    def __init__(self, base_url="http://127.0.0.1:8000/"):
        self._logger = logging.getLogger(Logger.client_logger_name)

        self.base_url = base_url

    def get(self, endpoint):
        """
        Used for get requests.
        """
        try:
            self._logger.info("<-- GET to %s%s", str(self.base_url), str(endpoint))
            url = self.base_url + endpoint
            with urlopen(url) as response:
                body = response.read()
                status = response.getcode()
            response_json = json.loads(body)
            self._logger.debug(response_json)
            return (status, response_json)
        except URLError:
            return (-1, "")
        except Exception:
            self._logger.error("Some other connection-error occured.")
            self._logger.error("<<<BEGIN-STACKTRACE>>>")
            traceback.print_exc()
            self._logger.error("<<<END-STACKTRACE>>>")
            return (-2, "")

    def post(self, endpoint, body: FrontendOperationDto):
        """
        Used for post requests.
        """
        try:
            self._logger.info(
                "<-- POST to %s%s   %s",
                str(self.base_url),
                str(endpoint),
                str(body.function),
            )

            data = body.to_JSON()
            data = data.encode("utf-8")
            req = Request(self.base_url + endpoint, data=data)
            req.add_header("Content-Type", "application/json")
            with urlopen(req) as response:
                body = response.read()
                status = response.getcode()
            response_json = json.loads(body)

            self._logger.debug(response_json)
            return (status, response_json)
        except URLError as e:
            self._logger.warning("Https-connect-error occured.")
            print(e)
            return (-1, "")
        except Exception:
            self._logger.error("Some other connection-error occured.")
            self._logger.error("<<<BEGIN-STACKTRACE>>>")
            traceback.print_exc()
            self._logger.error("<<<END-STACKTRACE>>>")
            return (-2, "")



"""
Module that contains the FontFamiliesType.
"""

from .type import Type
from .utils.standard_text_processing import StandardTextProcessing


class FontFamiliesType(Type):
    """
    The FontFamlies is a type, which can be used to specify the font family of a text.
    """

    @classmethod
    def parse(cls, parse_input: str, logger):
        parsed_string = StandardTextProcessing.parse_string_with_quotes(parse_input)

        return parsed_string

    @classmethod
    def description(cls):
        return (
            "As the font-families change from system to system, clinguin cannot provide a useful selection"
            + "of fonts here. Therefore one should search in one's favorite search engine, how to get a list"
            + "of available fonts for one's operating system. Generally one can then use the name of the"
            + "preferred font-family as the value. Some wide spread and popular font-families are"
            + "(NOTE THAT THEY MIGHT NOT WORK ON YOUR SYSTEM!!!): [''helvetica'',''times'',''arial'',''courier'']."
        )



# pylint: disable=R1732
"""
Wrapper module for starting the angular frontend.
"""

import http.server
import json
import os
import signal
import socketserver
import webbrowser
from typing import Any

from clinguin.client import AbstractFrontend

_ROOT = os.path.abspath(os.path.dirname(__file__))

HTML_FILES_RELATIVE_DIRECTORY = "clinguin_angular_frontend"
SERVED_DIRECTORY = os.path.join(
    _ROOT,
    HTML_FILES_RELATIVE_DIRECTORY,
)


class Handler(http.server.SimpleHTTPRequestHandler):
    """
    Handler for backend, it sets the server directory as ''SERVED_DIRECTORY'',
    which corresponds to ''clinguin_angular_frontend''.
    """

    # pylint: disable=redefined-builtin
    def __init__(self, *args, **kwargs):
        super().__init__(*args, directory=SERVED_DIRECTORY, **kwargs)

    def log_message(self, format: str, *args: Any) -> None:
        pass


class AngularFrontend(AbstractFrontend):
    """
    Class AngularFrontend extends like all other frontends the AbstractFrontend.
    It opens the angular server via ng serve.
    """

    @classmethod
    def register_options(cls, parser):
        parser.description = "This GUI is based on the Angular package."

        parser.add_argument(
            "--client-port",
            type=int,
            default=8087,
            help="Set the port for the webserver of the client.",
        )

    def __init__(self, base_engine, args):
        super().__init__(base_engine, args)
        if hasattr(args, "server_port"):
            server_port = args.server_port
        else:
            server_port = 8000

        if hasattr(args, "server_url"):
            server_url = args.server_url
        else:
            server_url = "http://localhost"

        config_dict = {}
        config_dict["serverPort"] = server_port
        config_dict["serverUrl"] = server_url

        with open(
            os.path.join(SERVED_DIRECTORY, "assets", "config.json"),
            "w",
            encoding="UTF-8",
        ) as config_file:
            config_file.write(str(json.dumps(config_dict)))

        socketserver.TCPServer.allow_reuse_address = True
        with socketserver.TCPServer(("", args.client_port), Handler) as httpd:
            print(
                f"\033[1;4m-> -> -> -> -> -> ->  Open the following link in your browser: \
                    \033[1;34m http://127.0.0.1:{args.client_port}\033[0m"
            )
            webbrowser.open(f"http://127.0.0.1:{args.client_port}", new=2)

            try:
                httpd.serve_forever()
            except KeyboardInterrupt:
                print("Interruption signal detected in frontend, shutting down!")
                try:
                    httpd.shutdown()
                    print("Frontend destruction successful!")
                except Exception:
                    print(
                        "Could not destroy frontend-server, frontend port could still be in use!"
                    )

                os.kill(os.getpid(), signal.SIGTERM)



"""
Clinguin Context passed to the clingo control object with helper python functions
"""

import hashlib
from clingo.symbol import String, SymbolType, Number
from clingo import parse_term


class ClinguinContext:
    """
    Makes available a set of python functions to be used in a UI encoding for handling strings.
    """

    def concat(self, *args):
        """
        Concatenates the given symbols as a string

        Example:
            .. code-block:: prolog

                attr(s_l(I), label, @concat("Semester ",I)):-semester(I).

            Label will be `Semester 1`
        Args:
            args: All symbols

        Returns:
            The string concatenating all symbols
        """
        new_atom = "".join([str(x).strip('"') for x in args])
        try:
            new_symbol = parse_term(new_atom)
        except Exception:
            new_symbol = String(new_atom)
        return new_symbol

    def format(self, s, *args):
        """
        Formats the string with the given arguments

        Example:
            .. code-block:: prolog

                attr(s_l(I), label, @format("Semester {}!",I)):-semester(I).

            Label will be `Semester 1!`

        Args:
            s (str): The string to format, for example "{0} and {1}"
            args: All symbols that can be accessed by the position starting in 0.
                If there is a single tuple as an argument, then its arguments are considered one by one.
        Returns:
            The string obtained by formatting the string with the given arguments
        """
        if (
            len(args) == 1
            and args[0].type == SymbolType.Function
            and args[0].name == ""
        ):
            args_str = [str(v).strip('"') for v in args[0].arguments]
        else:
            args_str = [str(v).strip('"') for v in args]
        new_atom = s.string.format(*args_str)
        try:
            new_symbol = parse_term(new_atom)
        except Exception:
            new_symbol = String(new_atom)
        return new_symbol

    def to_int(self, s):
        """
        Converts any symbol into an integer using a hash. It can be used to infer an order in the symbols
        for the `order` attribute in the UI.

        Example:
            .. code-block:: prolog

                attr(s_l(I), label, @to_int(person("Susana"))).

            Label will be `12`

        Args:
            s: The symbol to convert to an integer
        Returns:
            The integer value of the string
        """
        if s.type == SymbolType.Number:
            return s.number
        s = str(s)
        h = hashlib.sha256(s.encode()).hexdigest()
        return Number(int(h, 16) % 10**4)

    def replace(self, s, old, new):
        """
        Replaces all occurrences of the old string with the new string

        Example:
            .. code-block:: prolog

                attr(s_l(I), label, @replace("Semester 1", "1", "2")).

            Label will be `Semester 2`

        Args:
            s: The string to transform
            old: The string to replace
            new: The string to insert instead
        Returns:
            The string with the replaced values
        """
        val = str(s).strip('"')
        new_atom = val.replace(str(old).strip('"'), str(new).strip('"'))
        try:
            new_symbol = parse_term(new_atom)
        except Exception:
            new_symbol = String(new_atom)
        return new_symbol

    def stringify(self, s, capitalize=False):
        """
        Turns a value into a string without underscore and capitalized if requested

        Example:
            .. code-block:: prolog

                attr(s_l(I), label, @stringify(semester_1, true)). # Semester 1

            Label will be `Semester 1`

        Args:
            s: The value to transform
        Returns:
            The string without _
        """
        val = str(s).strip('"')
        val = val.replace("_", " ")
        if capitalize:
            val = val[0].upper() + val[1:]
        return String(val)

    def upper(self, s):
        """
        Turns a value into upper case

        Example:
            .. code-block:: prolog

                attr(s_l(I), label, @upper(semester_1)). # Semester 1

            Label will be `SEMESTER 1`

        Args:
            s: The value to transform
        Returns:
            The string without _
        """
        val = str(s).strip('"')
        return String(val.upper())

    def color(self, option, opacity=None):
        """
        Gets the html color code for the different options and the given opacity

        Args:
            option: primary, secondary, success, info, warning, danger, light
            opacity: Numeric value indicating the opacity of the color
        """
        option = str(option)
        opacity = str(opacity) if opacity is not None else None
        colors = {
            "primary": "#0052CC",
            "blue": "#0052CC",
            "secondary": "#6554C0",
            "purple": "#6554C0",
            "success": "#36B37E",
            "green": "#36B37E",
            "info": "#B3BAC5",
            "gray": "#B3BAC5",
            "warning": "#FFAB00",
            "yellow": "#FFAB00",
            "danger": "#FF5630",
            "red": "#FF5630",
            "light": "#F4F5F7",
        }
        if option not in colors:
            return String("#000000")

        hex_color = colors[option]

        if opacity is not None and opacity.isnumeric():
            o = int(opacity)
            if 0 <= o < 100:
                hex_color = f"{hex_color}{o:02d}"
        return String(hex_color)

    def decimal_string(self, number, decimals):
        """
        Converts a number given in percentage to a decimal string with the given number of decimals
        Args:
            number: The number in percentage
            decimals: The number of decimals to show
        Returns:
            The decimal string
        """
        number = number.number if number.type == SymbolType.Number else 0
        decimals = decimals.number if decimals.type == SymbolType.Number else 0

        decimal = number / 100
        return String(f"{decimal:.{decimals}f}")

    def __getattr__(self, name):
        # pylint: disable=import-outside-toplevel

        import __main__

        return getattr(__main__, name)



"""
Program base of client.
"""

from .api.api import Api
from .api.frontend_operation_dto import FrontendOperationDto
from .application.client_base import ClientBase
from .presentation.abstract_frontend import AbstractFrontend

__all__ = [
    AbstractFrontend.__name__,
    ClientBase.__name__,
    Api.__name__,
    FrontendOperationDto.__name__,
]



"""
This module contains the ExtensionClass
"""


class ExtensionClass:
    """
    If a class is a subtype of the ExtensionClass it resembles a set of attributes/callbacks
    that can be applied to elements.
    """

    @classmethod
    def get_attributes(cls, attributes=None):  # pylint: disable=W0613
        """
        Has to be implemented by children.
        """
        return {}

    @classmethod
    def get_callbacks(cls, attributes=None):  # pylint: disable=W0613
        """
        Has to be implemented by children.
        """
        return {}



"""
Module that contains the BackendOperationDto
"""

from typing import List, Optional

from pydantic import BaseModel


class ContextDto(BaseModel):
    """
    Optional pass to the backend, which handles the context.
    """

    key: str
    value: str


class BackendOperationDto(BaseModel):
    """
    Needed by the endpoints to get convert the transported json into something useful for the backend.
    """

    function: str
    context: Optional[List[ContextDto]] = []



"""
Module that contains the Clingcon Backend.
"""

import warnings

from clingo.ast import Location, Position, ProgramBuilder, Rule, parse_files
from clingcon import ClingconTheory


try:
    from fclingo import THEORY, Translator
    from fclingo.__main__ import DEF, Statistic
    from fclingo.parsing import HeadBodyTransformer

    _HAS_FCLINGO = True
except ImportError:
    _HAS_FCLINGO = False
    warnings.warn(
        "fclingo is not installed. The FclingoBackend will not work until you "
        "install it (e.g. `pip install git+https://github.com/potassco/fclingo.git`).",
        ImportWarning,
    )


from clinguin.server.application.backends.theory_backend import TheoryBackend
from clinguin.utils.annotations import extends


# pylint: disable=too-few-public-methods
class Config:
    """
    Configuration class for the fclingo backend.
    """

    def __init__(self, max_int, min_int, print_trans, defined) -> None:
        self.max_int = max_int
        self.min_int = min_int
        self.print_trans = print_trans
        self.defined = defined


class FclingoBackend(TheoryBackend):
    """
    Backend that allows programs using fclingo theory atoms as input.
    It also includes the assignment in the domain state.
    """

    theory_class = ClingconTheory

    def __init__(self, *args, **kwargs):
        if not _HAS_FCLINGO:
            raise RuntimeError(
                "fclingo is not installed. Please install it to use FclingoBackend.\n"
                "Try: pip install git+https://github.com/potassco/fclingo.git"
            )
        super().__init__(*args, **kwargs)

    @extends(TheoryBackend)
    def _load_file(self, f):
        """
        Uses the program builder to rewrite the theory atoms.
        """
        with ProgramBuilder(self._ctl) as bld:
            hbt = HeadBodyTransformer()  # type: ignore
            parse_files([f], lambda ast: bld.add(hbt.visit(ast)))
            pos = Position("<string>", 1, 1)
            loc = Location(pos, pos)
            for rule in hbt.rules_to_add:
                bld.add(Rule(loc, rule[0], rule[1]))

    @extends(TheoryBackend)
    def _ground(self, program="base", arguments=None):
        super()._ground(program, arguments)
        max_int = next((x[1] for x in self._theory_conf if x[0] == "max-int"), 1000)
        min_int = next((x[1] for x in self._theory_conf if x[0] == "min-int"), -1000)
        translator = Translator(  # type: ignore
            self._ctl, Config(max_int, min_int, False, DEF), Statistic()  # type: ignore
        )
        translator.translate(self._ctl.theory_atoms)

    @extends(TheoryBackend)
    def _create_ctl(self):
        """
        Registers the ClingconTheory.
        """
        super()._create_ctl()
        self._ctl.add("base", [], THEORY)  # type: ignore



"""
Module which contains the StandardJsonEncoder
"""

import logging

import networkx as nx

from clinguin.utils import Logger

from .attribute import AttributeDto
from .callback import CallbackDto
from .element import ElementDto


class StandardJsonEncoder:
    """
    The standrd json encoder is responsible for converting a UIState into a json hierarchy.
    """

    def __init__(self):
        pass

    @classmethod
    def encode(cls, ui_state, ds_dict):
        """
        Public easy to access wrapper for the _generate_hierarchy method.

        Arguments:
            ui_state : UIState
            ds_dict : Dict
        """
        elements_dict = {}

        root = ElementDto("root", "root", "root")
        elements_dict[str(root.id)] = root
        if not ds_dict:
            ds_dict = {}
        cls._generate_hierarchy(ui_state, root, elements_dict)
        return {"ui": root, "ds": ds_dict}

    @classmethod
    def _generate_hierarchy(cls, ui_state, hierarchy_root, elements_dict):
        """
        Converts the UIState into an Json Hierarchy (which is represented by an ElementDto).
        Therefore it first gets all dependencies, then orders the elements according to the dependencies
        and then adds for each element its attributes, callbacks and children.

        Arguments:
            ui_state : UIState
            hierarchy_root : ElementDto
            elements_dict : Dict
        """
        logger = logging.getLogger(Logger.server_logger_name)
        logger.debug("Generating JSON for UI")

        dependency = []
        elements_info = {}

        for w in ui_state.get_elements():
            elements_info[w.id] = {"parent": w.parent, "type": w.type}
            dependency.append((w.id, w.parent))
        for w in ui_state.get_elements():
            if str(w.parent) != "root" and w.parent not in elements_info:
                msg = "Error in %s. Parent element (ID: %s) undefined."
                logger.critical(
                    msg,
                    str(w),
                    str(w.parent),
                )
                raise Exception(msg % (str(w), str(w.parent)))

        directed_graph = nx.DiGraph(dependency)
        order = list(reversed(list(nx.topological_sort(directed_graph))))

        attrs = ui_state.get_attributes_grouped()
        attrs = {a[0]: list(a[1]) for a in attrs}
        cbs = ui_state.get_callbacks_grouped()
        cbs = {a[0]: list(a[1]) for a in cbs}

        for element_id in order:
            if str(element_id) == str(hierarchy_root.id):
                continue

            if element_id not in elements_info:
                logger.critical(
                    " The ID : '%s' was used as parent in an elem/3 predicate, could not be found!",
                    str(element_id),
                )
                raise Exception(" The ID : " + str(element_id) + " could not be found!")

            element_type = elements_info[element_id]["type"]
            parent = elements_info[element_id]["parent"]
            element = ElementDto(element_id, element_type, parent)

            if element_id in attrs:
                elem_attributes = [
                    AttributeDto(a.id, a.key, a.value) for a in attrs[element_id]
                ]
                element.set_attributes(elem_attributes)
            if element_id in cbs:
                elem_callbacks = [
                    CallbackDto(a.id, a.event, a.action, a.operation)
                    for a in cbs[element_id]
                ]
                element.set_callbacks(elem_callbacks)

            elements_dict[str(element_id)] = element
            elements_dict[str(element.parent)].add_child(element)



# pylint: disable=E1101
"""
This module contains the ConfigureSize class.
"""
from tkinter import font

from clinguin.utils.attribute_types import FontFamiliesType, FontWeightType, IntegerType

from .attribute_names import AttributeNames
from .extension_class import ExtensionClass


class ConfigureFont(ExtensionClass):
    """
    If a element is a subtype of the configure-size class the size of the element can be adjusted.
    It is required that any subtype of this type has the ''_configure_font_element'' variable
    (and must be set appropriately).
    """

    def __init__(self):
        self._configure_font_element = None

    @classmethod
    def get_attributes(cls, attributes=None):
        if attributes is None:
            attributes = {}

        attributes[AttributeNames.font_family] = {
            "value": "",
            "value_type": FontFamiliesType,
        }
        attributes[AttributeNames.font_size] = {"value": 12, "value_type": IntegerType}
        attributes[AttributeNames.font_weight] = {
            "value": "normal",
            "value_type": FontWeightType,
        }

        return attributes

    def _set_font(self, elements):  # pylint: disable=W0613
        family = family = self._attributes[AttributeNames.font_family]["value"]
        size = int(self._attributes[AttributeNames.font_size]["value"])
        weight = self._attributes[AttributeNames.font_weight]["value"]

        kwargs = {}
        if family != "":
            kwargs["family"] = family
        if size > 0:
            kwargs["size"] = size

        if weight in [FontWeightType.BOLD, FontWeightType.ITALIC_BOLD]:
            kwargs["weight"] = "bold"
        if weight in [FontWeightType.ITALIC, FontWeightType.ITALIC_BOLD]:
            kwargs["slant"] = "italic"

        if family != "" or size != "" or weight != FontWeightType.NORMAL:
            afont = font.Font(**kwargs)
            self._configure_font_element.configure(font=afont)
